from pathlib import Path

import pandas as pd
import pytest
from tomlkit import dumps, parse

from muse import examples
from muse.wizard import (
    add_agent,
    add_new_commodity,
    add_new_process,
    add_region,
    add_technodata_for_new_year,
    add_timeslice,
    get_sectors,
    modify_toml,
)


@pytest.fixture
def model_path(tmp_path):
    """Creates temporary folder containing the default model."""
    examples.copy_model(name="default", path=tmp_path)
    return tmp_path / "model"


@pytest.fixture
def model_path_retro(tmp_path):
    """Creates temporary folder containing the default_retro model."""
    examples.copy_model(name="default_retro", path=tmp_path)
    return tmp_path / "model"


def assert_values_in_csv(file_path: Path, column: str, expected_values: list):
    """Helper function to check if values exist in a CSV column."""
    df = pd.read_csv(file_path)
    for value in expected_values:
        assert value in df[column].values


def assert_columns_exist(file_path: Path, columns: list):
    """Helper function to check if columns exist in a CSV file."""
    df = pd.read_csv(file_path)
    for column in columns:
        assert column in df.columns


def test_modify_toml(tmp_path):
    """Test the modify_toml function."""
    toml_path = tmp_path / "temp.toml"
    initial_data = {"name": "Tomm", "age": 299}
    toml_path.write_text(dumps(initial_data))

    def modify_function(data):
        data.update({"name": "Tom", "age": 29})

    modify_toml(toml_path, modify_function)
    modified_data = parse(toml_path.read_text())

    assert modified_data == {"name": "Tom", "age": 29}


def test_get_sectors(tmp_path):
    """Test the get_sectors function."""
    model_path = tmp_path / "model"
    model_path.mkdir()

    # Create test sector folders
    for sector in ["sector1", "sector2", "sector3"]:
        sector_path = model_path / sector
        sector_path.mkdir(parents=True)
        if sector != "sector3":
            (sector_path / "Technodata.csv").touch()

    assert set(get_sectors(model_path)) == {"sector1", "sector2"}


def test_add_new_commodity(model_path):
    """Test the add_new_commodity function on the default model."""
    add_new_commodity(model_path, "new_commodity", "power", "wind")

    # Check global commodities
    assert_values_in_csv(
        model_path / "GlobalCommodities.csv", "commodity", ["new_commodity"]
    )


def test_add_new_process(model_path):
    """Test the add_new_process function on the default model."""
    add_new_process(model_path, "new_process", "power", "windturbine")

    files_to_check = [
        "CommIn.csv",
        "CommOut.csv",
        "ExistingCapacity.csv",
        "Technodata.csv",
    ]
    for file in files_to_check:
        assert_values_in_csv(model_path / "power" / file, "technology", ["new_process"])


def test_technodata_for_new_year(model_path):
    """Test the add_price_data_for_new_year function on the default model."""
    add_technodata_for_new_year(model_path, 2030, "power", 2020)

    assert_values_in_csv(model_path / "power" / "Technodata.csv", "year", [2030])


def test_add_agent(model_path_retro):
    """Test the add_agent function on the default_retro model."""
    add_agent(model_path_retro, "A2", "A1", "Agent3", "Agent4")

    # Check Agents.csv
    assert_values_in_csv(model_path_retro / "Agents.csv", "name", ["A2"])
    for share in ["Agent3", "Agent4"]:
        assert_values_in_csv(model_path_retro / "Agents.csv", "agent_share", [share])

    # Check Technodata.csv files
    for sector in ["power", "gas"]:
        assert_columns_exist(model_path_retro / sector / "Technodata.csv", ["Agent4"])


def test_add_region(model_path):
    """Test the add_region function on the default model."""
    add_region(model_path, "R2", "R1")

    # Check settings.toml
    with open(model_path / "settings.toml") as f:
        settings = parse(f.read())
        assert "R2" in settings["regions"]

    # Check sector files
    files_to_check = [
        "Technodata.csv",
        "CommIn.csv",
        "CommOut.csv",
        "ExistingCapacity.csv",
    ]
    for sector in get_sectors(model_path):
        for file in files_to_check:
            assert_values_in_csv(model_path / sector / file, "region", ["R2"])


def test_add_timeslice(model_path):
    """Test the add_timeslice function on the default model."""
    add_timeslice(model_path, "midnight", "evening")

    # Check settings.toml
    with open(model_path / "settings.toml") as f:
        settings = parse(f.read())
        timeslices = settings["timeslices"]["all-year"]["all-week"]
        assert "midnight" in timeslices
        n_timeslices = len(timeslices)

    # Check preset files
    for preset in ["Residential2020Consumption.csv", "Residential2050Consumption.csv"]:
        df = pd.read_csv(model_path / "residential_presets" / preset)
        assert len(df["timeslice"].unique()) == n_timeslices



"""Holds all building agents."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Callable

import xarray as xr

from muse.timeslices import drop_timeslice


class AbstractAgent(ABC):
    """Base class for all agents."""

    tolerance = 1e-12
    """tolerance criteria for floating point comparisons."""

    def __init__(
        self,
        name: str = "Agent",
        region: str = "",
        assets: xr.Dataset | None = None,
        category: str | None = None,
        quantity: float | None = 1,
        timeslice_level: str | None = None,
    ):
        """Creates a standard MUSE agent.

        Arguments:
            name: Name of the agent, used for cross-refencing external tables
            region: Region where the agent operates, used for cross-referencing
                external tables.
            assets: dataset holding information about the assets owned by this
                instance. The information should not be anything describing the
                technologies themselves, but rather the stock of assets held by
                the agent.
            category: optional value that could be used to classify different agents
                together.
            quantity: optional value to classify different agents' share of the
                population.
            timeslice_level: the timeslice level over which investments/production
                will be optimized (e.g "hour", "day"). If None, the agent will use the
                finest timeslice level.
        """
        from uuid import uuid4

        super().__init__()
        self.name = name
        """Name associated with the agent."""
        self.region = region
        """Region the agent operates in."""
        self.assets = assets if assets is not None else xr.Dataset()
        """Current stock of technologies."""
        self.uuid = uuid4()
        """A unique identifier for the agent."""
        self.category = category
        """Attribute to classify different sets of agents."""
        self.quantity = quantity
        """Attribute to classify different agents' share of the population."""
        self.timeslice_level = timeslice_level
        """Timeslice level for the agent."""

    def filter_input(
        self,
        dataset: xr.Dataset | xr.DataArray,
        **kwargs,
    ) -> xr.Dataset | xr.DataArray:
        """Filter inputs for usage in agent.

        For instance, filters down to agent's region, etc.
        """
        if "region" in dataset.dims and "region" not in kwargs:
            kwargs["region"] = self.region
        return dataset.sel(**kwargs)

    @abstractmethod
    def next(
        self,
        technologies: xr.Dataset,
        market: xr.Dataset,
        demand: xr.DataArray,
    ) -> None:
        """Increments agent to the next time point (e.g. performing investments).

        Performs investments to meet demands, and increments agent.year to the
        investment year.

        Arguments:
            technologies: dataset of technology parameters for the investment year
            market: market dataset covering the current year and investment year
            demand: data array of demand for the investment year
        """

    def __repr__(self):
        return (
            f"<{self.region}:({self.name}, {self.category}) "
            f"- {self.__class__.__name__} at "
            f"{hex(id(self))}>"
        )


class Agent(AbstractAgent):
    """Standard agent that does not perform investments."""

    def __init__(
        self,
        name: str = "Agent",
        region: str = "USA",
        assets: xr.Dataset | None = None,
        search_rules: Callable | None = None,
        objectives: Callable | None = None,
        decision: Callable | None = None,
        year: int = 2010,
        maturity_threshold: float = 0,
        housekeeping: Callable | None = None,
        merge_transform: Callable | None = None,
        demand_threshold: float | None = None,
        category: str | None = None,
        asset_threshold: float = 1e-4,
        quantity: float | None = 1,
        spend_limit: int = 0,
        timeslice_level: str | None = None,
        **kwargs,
    ):
        """Creates a standard agent.

        Arguments:
            name: Name of the agent, used for cross-refencing external tables
            region: Region where the agent operates, used for cross-referencing
                external tables.
            assets: Current stock of technologies.
            search_rules: method used to filter the search space
            objectives: One or more objectives by which to decide next investments.
            decision: single decision objective from one or more objectives.
            year: year the agent is created / current year
            maturity_threshold: threshold when filtering replacement
                technologies with respect to market share
            housekeeping: transform applied to the assets at the start of
                iteration. Defaults to doing nothing.
            merge_transform: transform merging current and newly invested assets
                together. Defaults to replacing old assets completely.
            demand_threshold: criteria below which the demand is zero.
            category: optional attribute that could be used to classify
                different agents together.
            asset_threshold: Threshold below which assets are not added.
            quantity: different agents' share of the population
            spend_limit: The cost above which agents will not invest
            timeslice_level: the timeslice level over which the agent invesments will
                be optimized (e.g "hour", "day"). If None, the agent will use the finest
                timeslice level.
            **kwargs: Extra arguments
        """
        from muse.decisions import factory as decision_factory
        from muse.filters import factory as filter_factory
        from muse.hooks import asset_merge_factory, housekeeping_factory
        from muse.objectives import factory as objectives_factory

        super().__init__(
            name=name,
            region=region,
            assets=assets,
            category=category,
            quantity=quantity,
            timeslice_level=timeslice_level,
        )

        """ Current year. Incremented by one every time next is called."""
        self.year = year

        """Search rule(s) determining potential replacement technologies.

        This is a string referring to a filter, or a sequence of strings
        referring to multiple filters, applied one after the other.  Any
        function registered via `muse.filters.register_filter` can be
        used to filter the search space.
        """
        if search_rules is None:
            search_rules = filter_factory()
        self.search_rules: Callable = search_rules

        """ Market share threshold.

        Threshold when and if filtering replacement technologies with respect
        to market share.
        """
        self.maturity_threshold = maturity_threshold

        self.spend_limit = spend_limit

        """One or more objectives by which to decide next investments."""
        if objectives is None:
            objectives = objectives_factory()
        self.objectives = objectives

        """Creates single decision objective from one or more objectives."""
        if decision is None:
            decision = decision_factory()
        self.decision = decision

        """Transforms applied on the assets at the start of each iteration.

        It could mean keeping the assets as are, or removing assets with no
        capacity in the current year and beyond, etc...
        It can be any function registered with
        :py:func:`~muse.hooks.register_initial_asset_transform`.
        """
        if housekeeping is None:
            housekeeping = housekeeping_factory()
        self._housekeeping = housekeeping

        """Transforms applied on the old and new assets.

        It could mean using only the new assets, or merging old and new, etc...
        It can be any function registered with
        :py:func:`~muse.hooks.register_final_asset_transform`.
        """
        if merge_transform is None:
            merge_transform = asset_merge_factory()
        self.merge_transform = merge_transform

        """Threshold below which the demand share is zero.

        This criteria avoids fulfilling demand for very small values. If None,
        then the criteria is not applied.
        """
        self.demand_threshold = demand_threshold

        """Threshold below which assets are not added."""
        self.asset_threshold = asset_threshold

    def asset_housekeeping(self):
        """Reduces memory footprint of assets.

        Performs tasks such as:

        - remove empty assets
        - remove years prior to current
        """
        # TODO: move this into search and make sure filters, demand_share and
        #  what not use assets from search. That would remove another bit of
        #  state.
        self.assets = self._housekeeping(self, self.assets)

    def next(
        self,
        technologies: xr.Dataset,
        market: xr.Dataset,
        demand: xr.DataArray,
    ) -> None:
        investment_year = int(market.year[1])
        self.year = investment_year


class InvestingAgent(Agent):
    """Agent that performs investment for itself."""

    def __init__(
        self,
        *args,
        constraints: Callable | None = None,
        investment: Callable | None = None,
        **kwargs,
    ):
        """Creates an investing agent.

        Arguments:
            *args: See :py:class:`~muse.agents.agent.Agent`
            constraints: Set of constraints limiting investment
            investment: A function to perform investments
            **kwargs: See :py:class:`~muse.agents.agent.Agent`
        """
        from muse.constraints import factory as csfactory
        from muse.investments import factory as ifactory

        super().__init__(*args, **kwargs)

        self.invest = investment or ifactory()
        """Method to use when fulfilling demand from rated set of techs."""
        self.constraints = constraints or csfactory()
        """Creates a set of constraints limiting investment."""

    def next(
        self,
        technologies: xr.Dataset,
        market: xr.Dataset,
        demand: xr.DataArray,
    ) -> None:
        """Iterates agent one turn.

        The goal is to figure out from market variables which technologies to
        invest in and by how much.

        This function will modify `self.assets` and increment `self.year`.
        Other attributes are left unchanged. Arguments to the function are
        never modified.
        """
        from logging import getLogger

        from muse.utilities import interpolate_capacity, reduce_assets

        # Check inputs
        assert len(market.year) == 2
        assert "year" not in technologies.dims
        assert "year" not in demand.dims

        # Time period
        current_year, investment_year = map(int, market.year.values)
        assert current_year == self.year

        # Skip forward if demand is zero
        if demand.size == 0 or demand.sum() < 1e-12:
            self.year = investment_year
            return None

        # Calculate the search space
        search_space = (
            self.search_rules(self, demand, technologies=technologies, market=market)
            .fillna(0)
            .astype(int)
        )

        # Select technologies in the search space
        technologies = technologies.sel(
            technology=technologies.technology.isin(search_space.replacement)
        )

        # Skip forward if the search space is empty
        if any(u == 0 for u in search_space.shape):
            getLogger(__name__).critical("Search space is empty")
            self.year = investment_year
            return None

        # Calculate the decision metric
        decision = self.compute_decision(technologies, market, demand, search_space)
        search = xr.Dataset(dict(search_space=search_space, decision=decision))
        if "timeslice" in search.dims:
            search["demand"] = drop_timeslice(demand)
        else:
            search["demand"] = demand

        # Filter assets with demand
        not_assets = [u for u in search.demand.dims if u != "asset"]
        condtechs = (
            search.demand.sum(not_assets) > getattr(self, "tolerance", 1e-8)
        ).values
        search = search.sel(asset=condtechs)

        # Calculate capacity in current and investment year
        capacity = interpolate_capacity(
            reduce_assets(self.assets.capacity, coords=("technology", "region")),
            year=[current_year, investment_year],
        )

        # Calculate constraints
        constraints = self.constraints(
            demand=search.demand,
            capacity=capacity,
            search_space=search.search_space,
            technologies=technologies,
            timeslice_level=self.timeslice_level,
        )

        # Calculate investments
        investments = self.invest(
            search=search[["search_space", "decision"]],
            technologies=technologies,
            constraints=constraints,
            commodities=list(demand.commodity.values),
            timeslice_level=self.timeslice_level,
        )

        # Add investments
        self.add_investments(
            technologies=technologies,
            investments=investments,
            investment_year=investment_year,
        )

        # Increment the year
        self.year = investment_year

    def compute_decision(
        self,
        technologies: xr.Dataset,
        market: xr.Dataset,
        demand: xr.DataArray,
        search_space: xr.DataArray,
    ) -> xr.DataArray:
        # Check inputs
        assert "year" not in technologies.dims
        assert "year" not in demand.dims
        assert "year" not in search_space.dims
        assert len(market.year) == 2

        # Filter technologies according to the search space and region
        techs = self.filter_input(
            technologies,
            technology=search_space.replacement,
        ).drop_vars("technology")

        # Reduce dimensions of the demand array
        reduced_demand = demand.sel(
            {
                k: search_space[k]
                for k in set(demand.dims).intersection(search_space.dims)
            }
        )

        # Filter prices according to the region
        prices = self.filter_input(market.prices)

        # Select prices for the investment year
        investment_year_prices = prices.isel(year=1)

        # Compute the objectives
        objectives = self.objectives(
            technologies=techs,
            demand=reduced_demand,
            prices=investment_year_prices,
            timeslice_level=self.timeslice_level,
        )

        # Compute the decision metric
        decision = self.decision(objectives)
        return decision

    def add_investments(
        self,
        technologies: xr.Dataset,
        investments: xr.DataArray,
        investment_year: int,
    ) -> None:
        """Add new assets to the agent."""
        assert "year" not in technologies.dims

        # Calculate retirement profile of new assets
        new_capacity = self.retirement_profile(
            technologies, investments, investment_year
        )
        if new_capacity is None:
            return
        new_capacity = new_capacity.drop_vars(
            set(new_capacity.coords) - set(self.assets.coords)
        )
        new_assets = xr.Dataset(dict(capacity=new_capacity))

        # Merge new assets with existing assets
        self.assets = self.merge_transform(self.assets, new_assets)

    def retirement_profile(
        self,
        technologies: xr.Dataset,
        investments: xr.DataArray,
        investment_year: int,
    ) -> xr.DataArray | None:
        from muse.investments import cliff_retirement_profile

        assert "year" not in technologies.dims

        # Sum investments
        if "asset" in investments.dims:
            investments = investments.sum("asset")
        if "agent" in investments.dims:
            investments = investments.squeeze("agent", drop=True)

        # Filter out investments below the threshold
        investments = investments.sel(
            replacement=(investments > self.asset_threshold).any(
                [d for d in investments.dims if d != "replacement"]
            )
        )
        if investments.size == 0:
            return None

        # Calculate the retirement profile for new investments
        # Note: technical life must be at least the length of the time period
        lifetime = self.filter_input(
            technologies.technical_life,
            technology=investments.replacement,
        )
        profile = cliff_retirement_profile(
            lifetime,
            investment_year=investment_year,
        )
        if "dst_region" in investments.coords:
            investments = investments.reindex_like(profile, method="ffill")

        # Apply the retirement profile to the investments
        new_assets = (investments * profile).rename(replacement="asset")
        new_assets["installed"] = "asset", [investment_year] * len(new_assets.asset)

        # The new assets have picked up quite a few coordinates along the way.
        # we try and keep only those that were there originally.
        if set(new_assets.dims) != set(self.assets.dims):
            new, old = new_assets.dims, self.assets.dims
            raise RuntimeError(f"Asset dimensions do not match: {new} vs {old}")
        return new_assets



"""Ensemble of functions to read MUSE data.

In general, there are three functions per input file:

- ``read_x``: This is the overall function that is called to read the data. It takes a
    ``Path`` as input, and returns the relevant data structure (usually an xarray). The
    process is generally broken down into two functions that are called by ``read_x``:

- ``read_x_csv``: This takes a path to a csv file as input and returns a pandas
    DataFrame. There are some consistency checks, such as checking data types and
    columns. There is also some minor processing at this stage, such as standardising
    column names, but no structural changes to the data. The general rule is that
    anything returned by this function should still be valid as an input file if saved
    to CSV.

- ``process_x``: This is where more major processing and reformatting of the data is
    done. It takes the DataFrame from ``read_x_csv`` and returns the final data
    structure (usually an xarray). There are also some more checks (e.g. checking for
    NaN values).

Most of the processing is shared by a few helper functions:
- read_csv: reads a csv file and returns a dataframe
- standardize_dataframe: standardizes the dataframe to a common format
- create_multiindex: creates a multiindex from a dataframe
- create_xarray_dataset: creates an xarray dataset from a dataframe

A few other helpers perform common operations on xarrays:
- create_assets: creates assets from technologies
- check_commodities: checks commodities and fills missing values

"""

from __future__ import annotations

__all__ = [
    "read_agent_parameters",
    "read_attribute_table",
    "read_csv",
    "read_existing_trade",
    "read_global_commodities",
    "read_initial_capacity",
    "read_initial_market",
    "read_io_technodata",
    "read_macro_drivers",
    "read_presets",
    "read_regression_parameters",
    "read_technodata_timeslices",
    "read_technodictionary",
    "read_technologies",
    "read_timeslice_shares",
    "read_trade_technodata",
]

from logging import getLogger
from pathlib import Path

import pandas as pd
import xarray as xr

from muse.utilities import camel_to_snake

# Global mapping of column names to their standardized versions
# This is for backwards compatibility with old file formats
COLUMN_RENAMES = {
    "process_name": "technology",
    "process": "technology",
    "sector_name": "sector",
    "region_name": "region",
    "time": "year",
    "commodity_name": "commodity",
    "comm_type": "commodity_type",
    "commodity_price": "prices",
    "units_commodity_price": "units_prices",
    "enduse": "end_use",
    "sn": "timeslice",
    "commodity_emission_factor_CO2": "emmission_factor",
    "utilisation_factor": "utilization_factor",
    "objsort": "obj_sort",
    "objsort1": "obj_sort1",
    "objsort2": "obj_sort2",
    "objsort3": "obj_sort3",
    "time_slice": "timeslice",
    "price": "prices",
}

# Columns who's values should be converted from camelCase to snake_case
CAMEL_TO_SNAKE_COLUMNS = [
    "tech_type",
    "commodity",
    "commodity_type",
    "agent_share",
    "attribute",
    "sector",
    "region",
    "parameter",
]

# Global mapping of column names to their expected types
COLUMN_TYPES = {
    "year": int,
    "region": str,
    "technology": str,
    "commodity": str,
    "sector": str,
    "attribute": str,
    "variable": str,
    "timeslice": int,  # For tables that require int timeslice instead of month etc.
    "name": str,
    "commodity_type": str,
    "tech_type": str,
    "type": str,
    "function_type": str,
    "level": str,
    "search_rule": str,
    "decision_method": str,
    "quantity": float,
    "share": float,
    "coeff": str,
    "value": float,
    "utilization_factor": float,
    "minimum_service_factor": float,
    "maturity_threshold": float,
    "spend_limit": float,
    "prices": float,
    "emmission_factor": float,
}

DEFAULTS = {
    "cap_par": 0,
    "cap_exp": 1,
    "fix_par": 0,
    "fix_exp": 1,
    "var_par": 0,
    "var_exp": 1,
    "interest_rate": 0,
    "utilization_factor": 1,
    "minimum_service_factor": 0,
    "search_rule": "all",
    "decision_method": "single",
}


def standardize_columns(data: pd.DataFrame) -> pd.DataFrame:
    """Standardizes column names in a DataFrame.

    This function:
    1. Converts column names to snake_case
    2. Applies the global COLUMN_RENAMES mapping
    3. Preserves any columns not in the mapping

    Args:
        data: DataFrame to standardize

    Returns:
        DataFrame with standardized column names
    """
    # Drop index column if present
    if data.columns[0] == "" or data.columns[0].startswith("Unnamed"):
        data = data.iloc[:, 1:]

    # Convert columns to snake_case
    data = data.rename(columns=camel_to_snake)

    # Then apply global mapping
    data = data.rename(columns=COLUMN_RENAMES)

    # Make sure there are no duplicate columns
    if len(data.columns) != len(set(data.columns)):
        raise ValueError(f"Duplicate columns in {data.columns}")

    return data


def create_multiindex(
    data: pd.DataFrame,
    index_columns: list[str],
    index_names: list[str],
    drop_columns: bool = True,
) -> pd.DataFrame:
    """Creates a MultiIndex from specified columns.

    Args:
        data: DataFrame to create index from
        index_columns: List of column names to use for index
        index_names: List of names for the index levels
        drop_columns: Whether to drop the original columns

    Returns:
        DataFrame with new MultiIndex
    """
    index = pd.MultiIndex.from_arrays(
        [data[col] for col in index_columns], names=index_names
    )
    result = data.copy()
    result.index = index
    if drop_columns:
        result = result.drop(columns=index_columns)
    return result


def create_xarray_dataset(
    data: pd.DataFrame,
    disallow_nan: bool = True,
) -> xr.Dataset:
    """Creates an xarray Dataset from a DataFrame with standardized options.

    Args:
        data: DataFrame to convert
        disallow_nan: Whether to raise an error if NaN values are found

    Returns:
        xarray Dataset
    """
    result = xr.Dataset.from_dataframe(data)
    if disallow_nan:
        nan_coords = get_nan_coordinates(result)
        if nan_coords:
            raise ValueError(f"Missing data for coordinates: {nan_coords}")

    if "year" in result.coords:
        result = result.assign_coords(year=result.year.astype(int))
        result = result.sortby("year")
        assert len(set(result.year.values)) == result.year.data.size  # no duplicates

    return result


def get_nan_coordinates(dataset: xr.Dataset) -> list[tuple]:
    """Get coordinates of a Dataset where any data variable has NaN values."""
    any_nan = sum(var.isnull() for var in dataset.data_vars.values())
    if any_nan.any():
        return any_nan.where(any_nan, drop=True).to_dataframe(name="").index.to_list()
    return []


def convert_column_types(data: pd.DataFrame) -> pd.DataFrame:
    """Converts DataFrame columns to their expected types.

    Args:
        data: DataFrame to convert

    Returns:
        DataFrame with converted column types
    """
    result = data.copy()
    for column, expected_type in COLUMN_TYPES.items():
        if column in result.columns:
            try:
                if expected_type is int:
                    result[column] = pd.to_numeric(result[column], downcast="integer")
                elif expected_type is float:
                    result[column] = pd.to_numeric(result[column]).astype(float)
                elif expected_type is str:
                    result[column] = result[column].astype(str)
            except (ValueError, TypeError) as e:
                raise ValueError(
                    f"Could not convert column '{column}' to {expected_type.__name__}: {e}"  # noqa: E501
                )
    return result


def standardize_dataframe(
    data: pd.DataFrame,
    required_columns: list[str] | None = None,
    exclude_extra_columns: bool = False,
) -> pd.DataFrame:
    """Standardizes a DataFrame to a common format.

    Args:
        data: DataFrame to standardize
        required_columns: List of column names that must be present (optional)
        exclude_extra_columns: If True, exclude any columns not in required_columns list
            (optional). This can be important if extra columns can mess up the resulting
            xarray object.

    Returns:
        DataFrame containing the standardized data
    """
    if required_columns is None:
        required_columns = []

    # Standardize column names
    data = standardize_columns(data)

    # Convert specified column values from camelCase to snake_case
    for col in CAMEL_TO_SNAKE_COLUMNS:
        if col in data.columns:
            data[col] = data[col].apply(camel_to_snake)

    # Fill missing values with defaults
    data = data.fillna(DEFAULTS)
    for col, default in DEFAULTS.items():
        if col not in data.columns and col in required_columns:
            data[col] = default

    # Check/convert data types
    data = convert_column_types(data)

    # Validate required columns if provided
    if required_columns:
        missing_columns = [col for col in required_columns if col not in data.columns]
        if missing_columns:
            raise ValueError(f"Missing required columns: {missing_columns}")

        # Exclude extra columns if requested
        if exclude_extra_columns:
            data = data[list(required_columns)]

    return data


def read_csv(
    path: Path,
    float_precision: str = "high",
    required_columns: list[str] | None = None,
    exclude_extra_columns: bool = False,
    msg: str | None = None,
) -> pd.DataFrame:
    """Reads and standardizes a CSV file into a DataFrame.

    Args:
        path: Path to the CSV file
        float_precision: Precision to use when reading floats
        required_columns: List of column names that must be present (optional)
        exclude_extra_columns: If True, exclude any columns not in required_columns list
            (optional). This can be important if extra columns can mess up the resulting
            xarray object.
        msg: Message to log (optional)

    Returns:
        DataFrame containing the standardized data
    """
    # Log message
    if msg:
        getLogger(__name__).info(msg)

    # Check if file exists
    if not path.is_file():
        raise OSError(f"{path} does not exist.")

    # Check if there's a units row (in which case we need to skip it)
    with open(path) as f:
        next(f)  # Skip header row
        first_data_row = f.readline().strip()
    skiprows = [1] if first_data_row.startswith("Unit") else None

    # Read the file
    data = pd.read_csv(
        path,
        float_precision=float_precision,
        low_memory=False,
        skiprows=skiprows,
    )

    # Standardize the DataFrame
    return standardize_dataframe(
        data,
        required_columns=required_columns,
        exclude_extra_columns=exclude_extra_columns,
    )


def check_commodities(
    data: xr.Dataset | xr.DataArray, fill_missing: bool = True, fill_value: float = 0
) -> xr.Dataset | xr.DataArray:
    """Validates and optionally fills missing commodities in data."""
    from muse.commodities import COMMODITIES

    # Make sure there are no commodities in data but not in global commodities
    extra_commodities = [
        c for c in data.commodity.values if c not in COMMODITIES.commodity.values
    ]
    if extra_commodities:
        raise ValueError(
            "The following commodities were not found in global commodities file: "
            f"{extra_commodities}"
        )

    # Add any missing commodities with fill_value
    if fill_missing:
        data = data.reindex(
            commodity=COMMODITIES.commodity.values, fill_value=fill_value
        )
    return data


def create_assets(data: xr.DataArray | xr.Dataset) -> xr.DataArray | xr.Dataset:
    """Creates assets from technology data."""
    # Rename technology to asset
    result = data.drop_vars("technology").rename(technology="asset")
    result["technology"] = "asset", data.technology.values

    # Add installed year
    result["installed"] = ("asset", [int(result.year.min())] * len(result.technology))
    return result


def read_technodictionary(path: Path) -> xr.Dataset:
    """Reads and processes technodictionary data from a CSV file."""
    df = read_technodictionary_csv(path)
    return process_technodictionary(df)


def read_technodictionary_csv(path: Path) -> pd.DataFrame:
    """Reads and formats technodata into a DataFrame."""
    required_columns = {
        "cap_exp",
        "region",
        "var_par",
        "fix_exp",
        "interest_rate",
        "utilization_factor",
        "minimum_service_factor",
        "year",
        "cap_par",
        "var_exp",
        "technology",
        "technical_life",
        "fix_par",
    }
    data = read_csv(
        path,
        required_columns=required_columns,
        msg=f"Reading technodictionary from {path}.",
    )

    # Check for deprecated columns
    if "fuel" in data.columns:
        msg = (
            f"The 'fuel' column in {path} has been deprecated. "
            "This information is now determined from CommIn files. "
            "Please remove this column from your Technodata files."
        )
        getLogger(__name__).warning(msg)
    if "end_use" in data.columns:
        msg = (
            f"The 'end_use' column in {path} has been deprecated. "
            "This information is now determined from CommOut files. "
            "Please remove this column from your Technodata files."
        )
        getLogger(__name__).warning(msg)
    if "scaling_size" in data.columns:
        msg = (
            f"The 'scaling_size' column in {path} has been deprecated. "
            "Please remove this column from your Technodata files."
        )
        getLogger(__name__).warning(msg)

    return data


def process_technodictionary(data: pd.DataFrame) -> xr.Dataset:
    """Processes technodictionary DataFrame into an xarray Dataset."""
    # Create multiindex for technology and region
    data = create_multiindex(
        data,
        index_columns=["technology", "region", "year"],
        index_names=["technology", "region", "year"],
        drop_columns=True,
    )

    # Create dataset
    result = create_xarray_dataset(data)

    # Handle tech_type if present
    if "type" in result.variables:
        result["tech_type"] = result.type.isel(region=0, year=0)

    return result


def read_technodata_timeslices(path: Path) -> xr.Dataset:
    """Reads and processes technodata timeslices from a CSV file."""
    df = read_technodata_timeslices_csv(path)
    return process_technodata_timeslices(df)


def read_technodata_timeslices_csv(path: Path) -> pd.DataFrame:
    """Reads and formats technodata timeslices into a DataFrame."""
    from muse.timeslices import TIMESLICE

    timeslice_columns = set(TIMESLICE.coords["timeslice"].indexes["timeslice"].names)
    required_columns = {
        "utilization_factor",
        "technology",
        "minimum_service_factor",
        "region",
        "year",
    } | timeslice_columns
    return read_csv(
        path,
        required_columns=required_columns,
        exclude_extra_columns=True,
        msg=f"Reading technodata timeslices from {path}.",
    )


def process_technodata_timeslices(data: pd.DataFrame) -> xr.Dataset:
    """Processes technodata timeslices DataFrame into an xarray Dataset."""
    from muse.timeslices import TIMESLICE, sort_timeslices

    # Create multiindex for all columns except factor columns
    factor_columns = ["utilization_factor", "minimum_service_factor", "obj_sort"]
    index_columns = [col for col in data.columns if col not in factor_columns]
    data = create_multiindex(
        data,
        index_columns=index_columns,
        index_names=index_columns,
        drop_columns=True,
    )

    # Create dataset
    result = create_xarray_dataset(data)

    # Stack timeslice levels (month, day, hour) into a single timeslice dimension
    timeslice_levels = TIMESLICE.coords["timeslice"].indexes["timeslice"].names
    if all(level in result.dims for level in timeslice_levels):
        result = result.stack(timeslice=timeslice_levels)
    return sort_timeslices(result)


def read_io_technodata(path: Path) -> xr.Dataset:
    """Reads and processes input/output technodata from a CSV file."""
    df = read_io_technodata_csv(path)
    return process_io_technodata(df)


def read_io_technodata_csv(path: Path) -> pd.DataFrame:
    """Reads process inputs or outputs into a DataFrame."""
    data = read_csv(
        path,
        required_columns=["technology", "region", "year"],
        msg=f"Reading IO technodata from {path}.",
    )

    # Unspecified Level values default to "fixed"
    if "level" in data.columns:
        data["level"] = data["level"].fillna("fixed")
    else:
        # Particularly relevant to outputs files where the Level column is omitted by
        # default, as only "fixed" outputs are allowed.
        data["level"] = "fixed"

    return data


def process_io_technodata(data: pd.DataFrame) -> xr.Dataset:
    """Processes IO technodata DataFrame into an xarray Dataset."""
    from muse.commodities import COMMODITIES

    # Extract commodity columns
    commodities = [c for c in data.columns if c in COMMODITIES.commodity.values]

    # Convert commodity columns to long format (i.e. single "commodity" column)
    data = data.melt(
        id_vars=["technology", "region", "year", "level"],
        value_vars=commodities,
        var_name="commodity",
        value_name="value",
    )

    # Pivot data to create fixed and flexible columns
    data = data.pivot(
        index=["technology", "region", "year", "commodity"],
        columns="level",
        values="value",
    )

    # Create xarray dataset
    result = create_xarray_dataset(data)

    # Ensure both `fixed` and `flexible` inputs/outputs are defined. If only one is
    # defined in the input data, create the other as a zeros array with the same shape.
    has_fixed = "fixed" in result.data_vars
    has_flexible = "flexible" in result.data_vars
    if has_fixed and not has_flexible:
        result["flexible"] = xr.zeros_like(result.fixed).rename("flexible")
    elif has_flexible and not has_fixed:
        result["fixed"] = xr.zeros_like(result.flexible).rename("fixed")
    elif not has_fixed and not has_flexible:
        raise ValueError("Neither 'fixed' nor 'flexible' levels were found.")

    # Fill any NaNs with zero
    result["fixed"] = result.fixed.fillna(0)
    result["flexible"] = result.flexible.fillna(0)

    # Check commodities
    result = check_commodities(result, fill_missing=True, fill_value=0)
    return result


def read_technologies(
    technodata_path: Path,
    comm_out_path: Path,
    comm_in_path: Path,
    time_framework: list[int],
    interpolation_mode: str = "linear",
    technodata_timeslices_path: Path | None = None,
) -> xr.Dataset:
    """Reads and processes technology data from multiple CSV files.

    Will also interpolate data to the time framework if provided.

    Args:
        technodata_path: path to the technodata file
        comm_out_path: path to the comm_out file
        comm_in_path: path to the comm_in file
        time_framework: list of years to interpolate data to
        interpolation_mode: Interpolation mode to use
        technodata_timeslices_path: path to the technodata_timeslices file

    Returns:
        xr.Dataset: Dataset containing the processed technology data. Any fields
        that differ by year will contain a "year" dimension interpolated to the
        time framework. Other fields will not have a "year" dimension.
    """
    # Read all data
    technodata = read_technodictionary(technodata_path)
    comm_out = read_io_technodata(comm_out_path)
    comm_in = read_io_technodata(comm_in_path)
    technodata_timeslices = (
        read_technodata_timeslices(technodata_timeslices_path)
        if technodata_timeslices_path
        else None
    )

    # Assemble xarray Dataset
    return process_technologies(
        technodata,
        comm_out,
        comm_in,
        time_framework,
        interpolation_mode,
        technodata_timeslices,
    )


def process_technologies(
    technodata: xr.Dataset,
    comm_out: xr.Dataset,
    comm_in: xr.Dataset,
    time_framework: list[int],
    interpolation_mode: str = "linear",
    technodata_timeslices: xr.Dataset | None = None,
) -> xr.Dataset:
    """Processes technology data DataFrames into an xarray Dataset."""
    from muse.commodities import COMMODITIES, CommodityUsage
    from muse.timeslices import drop_timeslice
    from muse.utilities import interpolate_technodata

    # Process inputs/outputs
    ins = comm_in.rename(flexible="flexible_inputs", fixed="fixed_inputs")
    outs = comm_out.rename(flexible="flexible_outputs", fixed="fixed_outputs")

    # Legacy: Remove flexible outputs
    if not (outs["flexible_outputs"] == 0).all():
        raise ValueError(
            "'flexible' outputs are not permitted. All outputs must be 'fixed'"
        )
    outs = outs.drop_vars("flexible_outputs")

    # Collect all years from the time framework and data files
    time_framework = list(
        set(time_framework).union(
            technodata.year.values.tolist(),
            ins.year.values.tolist(),
            outs.year.values.tolist(),
            technodata_timeslices.year.values.tolist() if technodata_timeslices else [],
        )
    )

    # Interpolate data to match the time framework
    technodata = interpolate_technodata(technodata, time_framework, interpolation_mode)
    outs = interpolate_technodata(outs, time_framework, interpolation_mode)
    ins = interpolate_technodata(ins, time_framework, interpolation_mode)
    if technodata_timeslices:
        technodata_timeslices = interpolate_technodata(
            technodata_timeslices, time_framework, interpolation_mode
        )

    # Merge inputs/outputs with technodata
    technodata = technodata.merge(outs).merge(ins)

    # Merge technodata_timeslices if provided. This will prioritise values defined in
    # technodata_timeslices, and fallback to the non-timesliced technodata for any
    # values that are not defined in technodata_timeslices.
    if technodata_timeslices:
        technodata["utilization_factor"] = (
            technodata_timeslices.utilization_factor.combine_first(
                technodata.utilization_factor
            )
        )
        technodata["minimum_service_factor"] = drop_timeslice(
            technodata_timeslices.minimum_service_factor.combine_first(
                technodata.minimum_service_factor
            )
        )

    # Check commodities
    technodata = check_commodities(technodata, fill_missing=False)

    # Add info about commodities
    technodata = technodata.merge(COMMODITIES.sel(commodity=technodata.commodity))

    # Add commodity usage flags
    technodata["comm_usage"] = (
        "commodity",
        CommodityUsage.from_technologies(technodata).values,
    )
    technodata = technodata.drop_vars("commodity_type")

    # Check utilization and minimum service factors
    check_utilization_and_minimum_service_factors(technodata)

    return technodata


def read_initial_capacity(path: Path) -> xr.DataArray:
    """Reads and processes initial capacity data from a CSV file."""
    df = read_initial_capacity_csv(path)
    return process_initial_capacity(df)


def read_initial_capacity_csv(path: Path) -> pd.DataFrame:
    """Reads and formats data about initial capacity into a DataFrame."""
    required_columns = {
        "region",
        "technology",
    }
    return read_csv(
        path,
        required_columns=required_columns,
        msg=f"Reading initial capacity from {path}.",
    )


def process_initial_capacity(data: pd.DataFrame) -> xr.DataArray:
    """Processes initial capacity DataFrame into an xarray DataArray."""
    # Drop unit column if present
    if "unit" in data.columns:
        data = data.drop(columns=["unit"])

    # Select year columns
    year_columns = [col for col in data.columns if col.isdigit()]

    # Convert year columns to long format (i.e. single "year" column)
    data = data.melt(
        id_vars=["technology", "region"],
        value_vars=year_columns,
        var_name="year",
        value_name="value",
    )

    # Create multiindex for region, technology, and year
    data = create_multiindex(
        data,
        index_columns=["technology", "region", "year"],
        index_names=["technology", "region", "year"],
        drop_columns=True,
    )

    # Create Dataarray
    result = create_xarray_dataset(data).value.astype(float)

    # Create assets
    result = create_assets(result)
    return result


def read_global_commodities(path: Path) -> xr.Dataset:
    """Reads and processes global commodities data from a CSV file."""
    df = read_global_commodities_csv(path)
    return process_global_commodities(df)


def read_global_commodities_csv(path: Path) -> pd.DataFrame:
    """Reads commodities information from input into a DataFrame."""
    # Due to legacy reasons, users can supply both Commodity and CommodityName columns
    # In this case, we need to remove the Commodity column to avoid conflicts
    # This is fine because Commodity just contains a long description that isn't needed
    getLogger(__name__).info(f"Reading global commodities from {path}.")
    df = pd.read_csv(path)
    df = df.rename(columns=camel_to_snake)
    if "commodity" in df.columns and "commodity_name" in df.columns:
        df = df.drop(columns=["commodity"])

    required_columns = {
        "commodity",
        "commodity_type",
    }
    data = standardize_dataframe(
        df,
        required_columns=required_columns,
    )

    # Raise warning if units are not defined
    if "unit" not in data.columns:
        msg = (
            "No units defined for commodities. Please define units for all commodities "
            "in the global commodities file."
        )
        getLogger(__name__).warning(msg)

    return data


def process_global_commodities(data: pd.DataFrame) -> xr.Dataset:
    """Processes global commodities DataFrame into an xarray Dataset."""
    # Drop description column if present. It's useful to include in the file, but we
    # don't need it for the simulation.
    if "description" in data.columns:
        data = data.drop(columns=["description"])

    data.index = [u for u in data.commodity]
    data = data.drop("commodity", axis=1)
    data.index.name = "commodity"
    return create_xarray_dataset(data)


def read_agent_parameters(path: Path) -> pd.DataFrame:
    """Reads and processes agent parameters from a CSV file."""
    df = read_agent_parameters_csv(path)
    return process_agent_parameters(df)


def read_agent_parameters_csv(path: Path) -> pd.DataFrame:
    """Reads standard MUSE agent-declaration csv-files into a DataFrame."""
    required_columns = {
        "search_rule",
        "quantity",
        "region",
        "type",
        "name",
        "agent_share",
        "decision_method",
    }
    data = read_csv(
        path,
        required_columns=required_columns,
        msg=f"Reading agent parameters from {path}.",
    )

    # Check for deprecated retrofit agents
    if "type" in data.columns:
        retrofit_agents = data[data.type.str.lower().isin(["retrofit", "retro"])]
        if not retrofit_agents.empty:
            msg = (
                "Retrofit agents will be deprecated in a future release. "
                "Please modify your model to use only agents of the 'New' type."
            )
            getLogger(__name__).warning(msg)

    # Legacy: drop AgentNumber column
    if "agent_number" in data.columns:
        data = data.drop(["agent_number"], axis=1)

    # Check consistency of objectives data columns
    objectives = [col for col in data.columns if col.startswith("objective")]
    floats = [col for col in data.columns if col.startswith("obj_data")]
    sorting = [col for col in data.columns if col.startswith("obj_sort")]

    if len(objectives) != len(floats) or len(objectives) != len(sorting):
        raise ValueError(
            "Agent objective, obj_data, and obj_sort columns are inconsistent in "
            f"{path}"
        )

    return data


def process_agent_parameters(data: pd.DataFrame) -> list[dict]:
    """Processes agent parameters DataFrame into a list of agent dictionaries."""
    result = []
    for _, row in data.iterrows():
        # Get objectives data
        objectives = (
            row[[i.startswith("objective") for i in row.index]].dropna().to_list()
        )
        sorting = row[[i.startswith("obj_sort") for i in row.index]].dropna().to_list()
        floats = row[[i.startswith("obj_data") for i in row.index]].dropna().to_list()

        # Create decision parameters
        decision_params = list(zip(objectives, sorting, floats))

        agent_type = {
            "new": "newcapa",
            "newcapa": "newcapa",
            "retrofit": "retrofit",
            "retro": "retrofit",
            "agent": "agent",
            "default": "agent",
        }[getattr(row, "type", "agent").lower()]

        # Create agent data dictionary
        data = {
            "name": row["name"],
            "region": row.region,
            "objectives": objectives,
            "search_rules": row.search_rule,
            "decision": {"name": row.decision_method, "parameters": decision_params},
            "agent_type": agent_type,
            "quantity": row.quantity,
            "share": row.agent_share,
        }

        # Add optional parameters
        if hasattr(row, "maturity_threshold"):
            data["maturity_threshold"] = row.maturity_threshold
        if hasattr(row, "spend_limit"):
            data["spend_limit"] = row.spend_limit

        # Add agent data to result
        result.append(data)

    return result


def read_initial_market(
    projections_path: Path,
    base_year_import_path: Path | None = None,
    base_year_export_path: Path | None = None,
    currency: str | None = None,
) -> xr.Dataset:
    """Reads and processes initial market data.

    Args:
        projections_path: path to the projections file
        base_year_import_path: path to the base year import file (optional)
        base_year_export_path: path to the base year export file (optional)
        currency: currency string (e.g. "USD")

    Returns:
        xr.Dataset: Dataset containing initial market data.
    """
    # Read projections
    projections_df = read_projections_csv(projections_path)

    # Read base year export (optional)
    if base_year_export_path:
        export_df = read_csv(
            base_year_export_path,
            msg=f"Reading base year export from {base_year_export_path}.",
        )
    else:
        export_df = None

    # Read base year import (optional)
    if base_year_import_path:
        import_df = read_csv(
            base_year_import_path,
            msg=f"Reading base year import from {base_year_import_path}.",
        )
    else:
        import_df = None

    # Assemble into xarray Dataset
    result = process_initial_market(projections_df, import_df, export_df, currency)
    return result


def read_projections_csv(path: Path) -> pd.DataFrame:
    """Reads projections data from a CSV file."""
    required_columns = {
        "region",
        "attribute",
        "year",
    }
    projections_df = read_csv(
        path, required_columns=required_columns, msg=f"Reading projections from {path}."
    )
    return projections_df


def process_initial_market(
    projections_df: pd.DataFrame,
    import_df: pd.DataFrame | None,
    export_df: pd.DataFrame | None,
    currency: str | None = None,
) -> xr.Dataset:
    """Process market data DataFrames into an xarray Dataset.

    Args:
        projections_df: DataFrame containing projections data
        import_df: Optional DataFrame containing import data
        export_df: Optional DataFrame containing export data
        currency: Currency string (e.g. "USD")
    """
    from muse.commodities import COMMODITIES
    from muse.timeslices import broadcast_timeslice, distribute_timeslice

    # Process projections
    projections = process_attribute_table(projections_df).commodity_price.astype(
        "float64"
    )

    # Process optional trade data
    if export_df is not None:
        base_year_export = process_attribute_table(export_df).exports.astype("float64")
    else:
        base_year_export = xr.zeros_like(projections)

    if import_df is not None:
        base_year_import = process_attribute_table(import_df).imports.astype("float64")
    else:
        base_year_import = xr.zeros_like(projections)

    # Distribute data over timeslices
    projections = broadcast_timeslice(projections, level=None)
    base_year_export = distribute_timeslice(base_year_export, level=None)
    base_year_import = distribute_timeslice(base_year_import, level=None)

    # Assemble into xarray
    result = xr.Dataset(
        {
            "prices": projections,
            "exports": base_year_export,
            "imports": base_year_import,
            "static_trade": base_year_import - base_year_export,
        }
    )

    # Check commodities
    result = check_commodities(result, fill_missing=True, fill_value=0)

    # Add units_prices coordinate
    # Only added if the currency is specified and commodity units are defined
    if currency and "unit" in COMMODITIES.data_vars:
        units_prices = [
            f"{currency}/{COMMODITIES.sel(commodity=c).unit.item()}"
            for c in result.commodity.values
        ]
        result = result.assign_coords(units_prices=("commodity", units_prices))

    return result


def read_attribute_table(path: Path) -> xr.Dataset:
    """Reads and processes attribute table data from a CSV file."""
    df = read_attribute_table_csv(path)
    return process_attribute_table(df)


def read_attribute_table_csv(path: Path) -> pd.DataFrame:
    """Read a standard MUSE csv file for price projections into a DataFrame."""
    table = read_csv(
        path,
        required_columns=["region", "attribute", "year"],
        msg=f"Reading attribute table from {path}.",
    )
    return table


def process_attribute_table(data: pd.DataFrame) -> xr.Dataset:
    """Process attribute table DataFrame into an xarray Dataset."""
    # Extract commodity columns
    commodities = [
        col for col in data.columns if col not in ["region", "year", "attribute"]
    ]

    # Convert commodity columns to long format (i.e. single "commodity" column)
    data = data.melt(
        id_vars=["region", "year", "attribute"],
        value_vars=commodities,
        var_name="commodity",
        value_name="value",
    )

    # Pivot data over attributes
    data = data.pivot(
        index=["region", "year", "commodity"],
        columns="attribute",
        values="value",
    )

    # Create DataSet
    result = create_xarray_dataset(data)
    return result


def read_presets(presets_paths: Path) -> xr.Dataset:
    """Reads and processes preset data from multiple CSV files.

    Accepts a path pattern for presets files, e.g. `Path("path/to/*Consumption.csv")`.
    The file name of each file must contain a year (e.g. "2020Consumption.csv").
    """
    from glob import glob
    from re import match

    # Find all files matching the path pattern
    allfiles = [Path(p) for p in glob(str(presets_paths))]
    if len(allfiles) == 0:
        raise OSError(f"No files found with paths {presets_paths}")

    # Read all files
    datas: dict[int, pd.DataFrame] = {}
    for path in allfiles:
        # Extract year from filename
        reyear = match(r"\S*.(\d{4})\S*\.csv", path.name)
        if reyear is None:
            raise OSError(f"Unexpected filename {path.name}")
        year = int(reyear.group(1))
        if year in datas:
            raise OSError(f"Year f{year} was found twice")

        # Read data
        data = read_presets_csv(path)
        data["year"] = year
        datas[year] = data

    # Process data
    datas = process_presets(datas)
    return datas


def read_presets_csv(path: Path) -> pd.DataFrame:
    data = read_csv(
        path,
        required_columns=["region", "timeslice"],
        msg=f"Reading presets from {path}.",
    )

    # Legacy: drop technology column and sum data (PR #448)
    if "technology" in data.columns:
        getLogger(__name__).warning(
            f"The technology (or ProcessName) column in file {path} is "
            "deprecated. Data has been summed across technologies, and this column "
            "has been dropped."
        )
        data = (
            data.drop(columns=["technology"])
            .groupby(["region", "timeslice"])
            .sum()
            .reset_index()
        )

    return data


def process_presets(datas: dict[int, pd.DataFrame]) -> xr.Dataset:
    """Processes preset DataFrames into an xarray Dataset."""
    from muse.commodities import COMMODITIES
    from muse.timeslices import TIMESLICE

    # Combine into a single DataFrame
    data = pd.concat(datas.values())

    # Extract commodity columns
    commodities = [c for c in data.columns if c in COMMODITIES.commodity.values]

    # Convert commodity columns to long format (i.e. single "commodity" column)
    data = data.melt(
        id_vars=["region", "year", "timeslice"],
        value_vars=commodities,
        var_name="commodity",
        value_name="value",
    )

    # Create multiindex for region, year, timeslice and commodity
    data = create_multiindex(
        data,
        index_columns=["region", "year", "timeslice", "commodity"],
        index_names=["region", "year", "timeslice", "commodity"],
        drop_columns=True,
    )

    # Create DataArray
    result = create_xarray_dataset(data).value.astype(float)

    # Assign timeslices
    result = result.assign_coords(timeslice=TIMESLICE.timeslice)

    # Check commodities
    result = check_commodities(result, fill_missing=True, fill_value=0)
    return result


def read_trade_technodata(path: Path) -> xr.Dataset:
    """Reads and processes trade technodata from a CSV file."""
    df = read_trade_technodata_csv(path)
    return process_trade_technodata(df)


def read_trade_technodata_csv(path: Path) -> pd.DataFrame:
    required_columns = {"technology", "region", "parameter"}
    return read_csv(
        path,
        required_columns=required_columns,
        msg=f"Reading trade technodata from {path}.",
    )


def process_trade_technodata(data: pd.DataFrame) -> xr.Dataset:
    # Drop unit column if present
    if "unit" in data.columns:
        data = data.drop(columns=["unit"])

    # Select region columns
    # TODO: this is a bit unsafe as user could supply other columns
    regions = [
        col for col in data.columns if col not in ["technology", "region", "parameter"]
    ]

    # Melt data over regions
    data = data.melt(
        id_vars=["technology", "region", "parameter"],
        value_vars=regions,
        var_name="dst_region",
        value_name="value",
    )

    # Pivot data over parameters
    data = data.pivot(
        index=["technology", "region", "dst_region"],
        columns="parameter",
        values="value",
    )

    # Create DataSet
    return create_xarray_dataset(data)


def read_existing_trade(path: Path) -> xr.DataArray:
    """Reads and processes existing trade data from a CSV file."""
    df = read_existing_trade_csv(path)
    return process_existing_trade(df)


def read_existing_trade_csv(path: Path) -> pd.DataFrame:
    required_columns = {
        "region",
        "technology",
        "year",
    }
    return read_csv(
        path,
        required_columns=required_columns,
        msg=f"Reading existing trade from {path}.",
    )


def process_existing_trade(data: pd.DataFrame) -> xr.DataArray:
    # Select region columns
    # TODO: this is a bit unsafe as user could supply other columns
    regions = [
        col for col in data.columns if col not in ["technology", "region", "year"]
    ]

    # Melt data over regions
    data = data.melt(
        id_vars=["technology", "region", "year"],
        value_vars=regions,
        var_name="dst_region",
        value_name="value",
    )

    # Create multiindex for region, dst_region, technology and year
    data = create_multiindex(
        data,
        index_columns=["region", "dst_region", "technology", "year"],
        index_names=["region", "dst_region", "technology", "year"],
        drop_columns=True,
    )

    # Create DataArray
    result = create_xarray_dataset(data).value.astype(float)

    # Create assets from technologies
    result = create_assets(result)
    return result


def read_timeslice_shares(path: Path) -> xr.DataArray:
    """Reads and processes timeslice shares data from a CSV file."""
    df = read_timeslice_shares_csv(path)
    return process_timeslice_shares(df)


def read_timeslice_shares_csv(path: Path) -> pd.DataFrame:
    """Reads sliceshare information into a DataFrame."""
    data = read_csv(
        path,
        required_columns=["region", "timeslice"],
        msg=f"Reading timeslice shares from {path}.",
    )

    return data


def process_timeslice_shares(data: pd.DataFrame) -> xr.DataArray:
    """Processes timeslice shares DataFrame into an xarray DataArray."""
    from muse.commodities import COMMODITIES
    from muse.timeslices import TIMESLICE

    # Extract commodity columns
    commodities = [c for c in data.columns if c in COMMODITIES.commodity.values]

    # Convert commodity columns to long format (i.e. single "commodity" column)
    data = data.melt(
        id_vars=["region", "timeslice"],
        value_vars=commodities,
        var_name="commodity",
        value_name="value",
    )

    # Create multiindex for region and timeslice
    data = create_multiindex(
        data,
        index_columns=["region", "timeslice", "commodity"],
        index_names=["region", "timeslice", "commodity"],
        drop_columns=True,
    )

    # Create DataArray
    result = create_xarray_dataset(data).value.astype(float)

    # Assign timeslices
    result = result.assign_coords(timeslice=TIMESLICE.timeslice)

    # Check commodities
    result = check_commodities(result, fill_missing=True, fill_value=0)
    return result


def read_macro_drivers(path: Path) -> pd.DataFrame:
    """Reads and processes macro drivers data from a CSV file."""
    df = read_macro_drivers_csv(path)
    return process_macro_drivers(df)


def read_macro_drivers_csv(path: Path) -> pd.DataFrame:
    """Reads a standard MUSE csv file for macro drivers into a DataFrame."""
    table = read_csv(
        path,
        required_columns=["region", "variable"],
        msg=f"Reading macro drivers from {path}.",
    )

    # Validate required variables
    required_variables = ["Population", "GDP|PPP"]
    missing_variables = [
        var for var in required_variables if var not in table.variable.unique()
    ]
    if missing_variables:
        raise ValueError(f"Missing required variables in {path}: {missing_variables}")

    return table


def process_macro_drivers(data: pd.DataFrame) -> xr.Dataset:
    """Processes macro drivers DataFrame into an xarray Dataset."""
    # Drop unit column if present
    if "unit" in data.columns:
        data = data.drop(columns=["unit"])

    # Select year columns
    year_columns = [col for col in data.columns if col.isdigit()]

    # Convert year columns to long format (i.e. single "year" column)
    data = data.melt(
        id_vars=["variable", "region"],
        value_vars=year_columns,
        var_name="year",
        value_name="value",
    )

    # Pivot data to create Population and GDP|PPP columns
    data = data.pivot(
        index=["region", "year"],
        columns="variable",
        values="value",
    )

    # Legacy: rename Population to population and GDP|PPP to gdp
    if "Population" in data.columns:
        data = data.rename(columns={"Population": "population"})
    if "GDP|PPP" in data.columns:
        data = data.rename(columns={"GDP|PPP": "gdp"})

    # Create DataSet
    result = create_xarray_dataset(data)
    return result


def read_regression_parameters(path: Path) -> xr.Dataset:
    """Reads and processes regression parameters from a CSV file."""
    df = read_regression_parameters_csv(path)
    return process_regression_parameters(df)


def read_regression_parameters_csv(path: Path) -> pd.DataFrame:
    """Reads the regression parameters from a MUSE csv file into a DataFrame."""
    table = read_csv(
        path,
        required_columns=["region", "function_type", "coeff"],
        msg=f"Reading regression parameters from {path}.",
    )

    # Legacy: warn about "sector" column
    if "sector" in table.columns:
        getLogger(__name__).warning(
            f"The sector column (in file {path}) is deprecated. Please remove."
        )

    return table


def process_regression_parameters(data: pd.DataFrame) -> xr.Dataset:
    """Processes regression parameters DataFrame into an xarray Dataset."""
    from muse.commodities import COMMODITIES

    # Extract commodity columns
    commodities = [c for c in data.columns if c in COMMODITIES.commodity.values]

    # Melt to long format
    melted = data.melt(
        id_vars=["sector", "region", "function_type", "coeff"],
        value_vars=commodities,
        var_name="commodity",
        value_name="value",
    )

    # Extract sector -> function_type mapping
    sector_to_ftype = melted.drop_duplicates(["sector", "function_type"])[
        ["sector", "function_type"]
    ].set_index("sector")["function_type"]

    # Pivot to create coefficient variables
    pivoted = melted.pivot_table(
        index=["sector", "region", "commodity"], columns="coeff", values="value"
    )

    # Create dataset and add function_type
    result = create_xarray_dataset(pivoted)
    result["function_type"] = xr.DataArray(
        sector_to_ftype[result.sector.values].astype(object),
        dims=["sector"],
        name="function_type",
    )

    # Check commodities
    result = check_commodities(result, fill_missing=True, fill_value=0)
    return result


def check_utilization_and_minimum_service_factors(data: xr.Dataset) -> None:
    """Check utilization and minimum service factors in an xarray dataset.

    Args:
        data: xarray Dataset containing utilization_factor and minimum_service_factor
    """
    if "utilization_factor" not in data.data_vars:
        raise ValueError(
            "A technology needs to have a utilization factor defined for every "
            "timeslice."
        )

    # Check UF not all zero (sum across timeslice dimension if it exists)
    if "timeslice" in data.dims:
        utilization_sum = data.utilization_factor.sum(dim="timeslice")
    else:
        utilization_sum = data.utilization_factor

    if (utilization_sum == 0).any():
        raise ValueError(
            "A technology can not have a utilization factor of 0 for every timeslice."
        )

    # Check UF in range
    utilization = data.utilization_factor
    if not ((utilization >= 0) & (utilization <= 1)).all():
        raise ValueError(
            "Utilization factor values must all be between 0 and 1 inclusive."
        )

    # Check MSF in range
    min_service_factor = data.minimum_service_factor
    if not ((min_service_factor >= 0) & (min_service_factor <= 1)).all():
        raise ValueError(
            "Minimum service factor values must all be between 0 and 1 inclusive."
        )

    # Check UF not below MSF
    if (data.utilization_factor < data.minimum_service_factor).any():
        raise ValueError(
            "Utilization factors must all be greater than or equal "
            "to their corresponding minimum service factors."
        )



from itertools import chain, permutations
from pathlib import Path

import pandas as pd
import toml
import xarray as xr
from pytest import fixture, mark, raises


@fixture
def user_data_files(settings: dict) -> None:
    """Creates test files related to user data."""
    for file_key, file_path in settings["global_input_files"].items():
        if file_key == "path" or not file_path:
            continue
        path = Path(file_path)
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text("Some data")


@fixture
def plugins(settings: dict, tmp_path) -> Path:
    """Creates test files related to custom modules."""
    plugin = tmp_path / "plugins" / "cat.py"
    plugin.parent.mkdir(parents=True, exist_ok=True)
    plugin.write_text("my_cat_colour = 'tabby' ")
    settings["plugins"] = str(plugin)
    return plugin


def test_add_known_parameters(settings: dict):
    """Test adding known parameters to settings."""
    from muse.readers import DEFAULT_SETTINGS_PATH
    from muse.readers.toml import MissingSettings, add_known_parameters

    defaults = toml.load(DEFAULT_SETTINGS_PATH)
    settings_copy = settings.copy()

    # Test required parameters
    settings_copy.pop("time_framework")
    with raises(MissingSettings):
        add_known_parameters(defaults, settings_copy)

    # Test optional parameters
    settings_copy = settings.copy()
    settings_copy.pop("global_input_files")
    new_settings = add_known_parameters(defaults, settings_copy)
    assert "projections" in new_settings["global_input_files"]


def test_add_unknown_parameters(settings: dict):
    """Test adding unknown parameters to settings."""
    from muse.readers import DEFAULT_SETTINGS_PATH
    from muse.readers.toml import add_unknown_parameters

    defaults = toml.load(DEFAULT_SETTINGS_PATH)
    settings_copy = settings.copy()
    settings_copy["my_new_parameter"] = 42

    new_settings = add_unknown_parameters(defaults, settings_copy)
    assert "my_new_parameter" in new_settings


def test_check_log_level(settings: dict):
    """Tests log level validation."""
    from muse.readers.toml import check_log_level

    check_log_level(settings)


def test_check_interpolation_mode(settings: dict):
    """Tests interpolation mode validation."""
    from muse.readers.toml import check_interpolation_mode

    check_interpolation_mode(settings)


def test_check_budget_parameters(settings: dict):
    """Tests budget parameters validation."""
    from muse.readers.toml import check_budget_parameters

    check_budget_parameters(settings)


def test_check_plugins(settings: dict, plugins: Path):
    """Tests plugin validation."""
    from muse.readers.toml import IncorrectSettings, check_plugins

    check_plugins(settings)

    # Test plugin not found error
    settings["plugins"] = plugins.parent / f"{plugins.stem}_2{plugins.suffix}"
    with raises(IncorrectSettings):
        check_plugins(settings)


def test_check_iteration_control(settings: dict):
    """Tests iteration control validation."""
    from muse.readers.toml import check_iteration_control

    # Test valid settings
    settings.update({"equilibrium": "off", "maximum_iterations": 7})
    check_iteration_control(settings)
    assert not settings["equilibrium"]

    # Test invalid maximum iterations
    settings.update({"equilibrium": True, "maximum_iterations": -1})
    with raises(ValueError):
        check_iteration_control(settings)

    # Test invalid tolerance
    settings.update({"maximum_iterations": 5, "tolerance": -1})
    with raises(ValueError):
        check_iteration_control(settings)


def test_format_paths(tmp_path):
    """Test format_paths with various settings structures."""
    from muse.readers.toml import format_paths

    # Flat dict with path-like values
    settings = {
        "input_file": "{path}/data.csv",
        "script": "{cwd}/run.py",
        "not_a_path": "foo.txt",
        "filename": "{cwd}/special",
    }
    path = tmp_path
    cwd = tmp_path.parent
    result = format_paths(settings, path=path, cwd=cwd)
    assert result["input_file"] == (path / "data.csv").absolute()
    assert result["script"] == (cwd / "run.py").absolute()
    assert result["not_a_path"] == "foo.txt"
    assert result["filename"] == (cwd / "special").absolute()

    # Nested dict
    settings = {
        "level1": {
            "input_file": "{path}/nested.csv",
            "filename": "{cwd}/nested.py",
        },
        "other": 123,
    }
    result = format_paths(settings, path=path, cwd=cwd)
    assert result["level1"]["input_file"] == (path / "nested.csv").absolute()
    assert result["level1"]["filename"] == (cwd / "nested.py").absolute()
    assert result["other"] == 123

    # List of paths and non-paths
    settings = {
        "files": [
            "{path}/a.csv",
            "{cwd}/b.py",
            "not_a_path",
            {"input_file": "{cwd}/c.nc"},
        ]
    }
    result = format_paths(settings, path=path, cwd=cwd)
    assert result["files"][0] == (path / "a.csv").absolute()
    assert result["files"][1] == (cwd / "b.py").absolute()
    assert result["files"][2] == "not_a_path"
    assert result["files"][3]["input_file"] == (cwd / "c.nc").absolute()


@mark.parametrize("suffix", [".xlsx", ".csv", ".toml", ".py", ".xls", ".nc"])
def test_suffix_path_formatting(suffix, tmp_path):
    """Test path formatting with different file suffixes."""
    from muse.readers.toml import read_toml

    # Test path formatting
    settings = {"this": 0, "plugins": f"{{path}}/thisfile{suffix}"}
    input_file = tmp_path / "settings.toml"
    input_file.write_text(toml.dumps(settings), encoding="utf-8")

    result = read_toml(input_file, path=tmp_path)
    assert result["plugins"].resolve() == (tmp_path / f"thisfile{suffix}").resolve()

    # Test cwd formatting
    settings["plugins"] = [f"{{cwd}}/other/thisfile{suffix}"]
    input_file.write_text(toml.dumps(settings), encoding="utf-8")

    result = read_toml(input_file, path="hello")
    assert result["plugins"][0].resolve() == (
        (Path.cwd() / "other" / f"thisfile{suffix}").resolve()
    )


def test_check_utilization_and_minimum_service():
    """Test combined validation of utilization and minimum service factors."""
    from muse.readers.csv import check_utilization_and_minimum_service_factors

    # Test valid case - create dataset with proper dimensions
    ds = xr.Dataset(
        {
            "utilization_factor": xr.DataArray(
                [[0.5, 0.5], [0.3, 0.7]],
                dims=["technology", "timeslice"],
                coords={"technology": ["tech1", "tech2"], "timeslice": [1, 2]},
            ),
            "minimum_service_factor": xr.DataArray(
                [[0.1, 0.1], [0.0, 0.0]],
                dims=["technology", "timeslice"],
                coords={"technology": ["tech1", "tech2"], "timeslice": [1, 2]},
            ),
        }
    )
    check_utilization_and_minimum_service_factors(ds)

    # Test utilization below minimum
    ds = xr.Dataset(
        {
            "utilization_factor": xr.DataArray(
                [[0.5, 0.5], [0.3, 0.7]],
                dims=["technology", "timeslice"],
                coords={"technology": ["tech1", "tech2"], "timeslice": [1, 2]},
            ),
            "minimum_service_factor": xr.DataArray(
                [[0.6, 0.1], [0.0, 0.0]],
                dims=["technology", "timeslice"],
                coords={"technology": ["tech1", "tech2"], "timeslice": [1, 2]},
            ),
        }
    )
    with raises(ValueError):
        check_utilization_and_minimum_service_factors(ds)

    # Test missing utilization factor
    ds = xr.Dataset(
        {
            "technology": xr.DataArray(
                ["tech1", "tech2"],
                dims=["technology"],
                coords={"technology": ["tech1", "tech2"]},
            ),
        }
    )
    with raises(ValueError):
        check_utilization_and_minimum_service_factors(ds)


def test_check_utilization_not_all_zero_fail():
    """Test validation fails when all utilization factors are zero."""
    from muse.readers.csv import check_utilization_and_minimum_service_factors

    ds = xr.Dataset(
        {
            "utilization_factor": xr.DataArray(
                [[0.0, 0.0], [0.3, 0.7]],
                dims=["technology", "timeslice"],
                coords={"technology": ["tech1", "tech2"], "timeslice": [1, 2]},
            ),
        }
    )
    with raises(ValueError):
        check_utilization_and_minimum_service_factors(ds)


@mark.parametrize(
    "values", chain.from_iterable(permutations((0, bad)) for bad in (-1, 2))
)
def test_check_utilization_in_range_fail(values):
    """Test validation fails for utilization factors outside valid range."""
    from muse.readers.csv import check_utilization_and_minimum_service_factors

    ds = xr.Dataset(
        {
            "utilization_factor": xr.DataArray(
                list(values),
                dims=["timeslice"],
                coords={"timeslice": range(len(values))},
            ),
        }
    )
    with raises(ValueError):
        check_utilization_and_minimum_service_factors(ds)


def test_get_nan_coordinates():
    """Test get_nan_coordinates for various scenarios."""
    from muse.readers.csv import get_nan_coordinates

    # Test 1: Explicit NaN values
    df1 = pd.DataFrame(
        {
            "region": ["R1", "R1", "R2", "R2"],
            "year": [2020, 2021, 2020, 2021],
            "value": [1.0, float("nan"), 3.0, 4.0],
        }
    )
    dataset1 = xr.Dataset.from_dataframe(df1.set_index(["region", "year"]))
    nan_coords1 = get_nan_coordinates(dataset1)
    assert nan_coords1 == [("R1", 2021)]

    # Test 2: Missing coordinate combinations
    df2 = pd.DataFrame(
        {
            "region": ["R1", "R1", "R2"],  # Missing R2-2021
            "year": [2020, 2021, 2020],
            "value": [1.0, 2.0, 3.0],
        }
    )
    dataset2 = xr.Dataset.from_dataframe(df2.set_index(["region", "year"]))
    nan_coords2 = get_nan_coordinates(dataset2)
    assert nan_coords2 == [("R2", 2021)]

    # Test 3: No NaN values
    df3 = pd.DataFrame(
        {
            "region": ["R1", "R1", "R2", "R2"],
            "year": [2020, 2021, 2020, 2021],
            "value": [1.0, 2.0, 3.0, 4.0],
        }
    )
    dataset3 = xr.Dataset.from_dataframe(df3.set_index(["region", "year"]))
    nan_coords3 = get_nan_coordinates(dataset3)
    assert nan_coords3 == []



"""Timeslice utility functions."""

from __future__ import annotations

__all__ = [
    "broadcast_timeslice",
    "compress_timeslice",
    "distribute_timeslice",
    "drop_timeslice",
    "expand_timeslice",
    "get_level",
    "read_timeslices",
    "setup_module",
    "sort_timeslices",
    "timeslice_max",
]

from collections.abc import Mapping, Sequence

import numpy as np
import pandas as pd
from xarray import DataArray

TIMESLICE: DataArray = None  # type: ignore


def read_timeslices(
    settings: Mapping | str,
    level_names: Sequence[str] = ("month", "day", "hour"),
) -> DataArray:
    from functools import reduce
    from logging import getLogger

    from toml import loads

    # Read timeslice settings
    if isinstance(settings, str):
        settings = loads(settings)
    settings = dict(**settings.get("timeslices", settings))

    # Legacy: warn user about deprecation of "aggregates" feature (#550)
    if "aggregates" in settings:
        msg = (
            "Timeslice aggregation has been deprecated since v1.3.0. Please see the "
            "release notes for that version for more information."
        )
        getLogger(__name__).warning(msg)
        settings.pop("aggregates")

    # Extract level names
    if "level_names" in settings:
        level_names = settings.pop("level_names")

    # Extract timeslice levels and lengths
    ts = list(settings.values())
    levels: list[tuple] = [(level,) for level in settings]
    while all(isinstance(v, Mapping) for v in ts):
        levels = [(*previous, b) for previous, a in zip(levels, ts) for b in a]
        ts = reduce(list.__add__, (list(u.values()) for u in ts), [])

    # Prepare multiindex
    nln = min(len(levels[0]), len(level_names))
    level_names = (
        list(level_names[:nln]) + [str(i) for i in range(len(levels[0]))][nln:]
    )
    indices = pd.MultiIndex.from_tuples(levels, names=level_names)

    # Make sure names from different levels don't overlap
    if any(
        reduce(set.union, indices.levels[:i], set()).intersection(indices.levels[i])
        for i in range(1, indices.nlevels)
    ):
        raise ValueError("Names from different levels should not overlap.")

    # Create DataArray
    return DataArray(ts, coords={"timeslice": indices}, dims="timeslice")


def setup_module(settings: str | Mapping):
    """Sets up module singletons."""
    global TIMESLICE
    TIMESLICE = read_timeslices(settings)


def broadcast_timeslice(
    data: DataArray, ts: DataArray | None = None, level: str | None = None
) -> DataArray:
    """Convert a non-timesliced array to a timesliced array by broadcasting.

    If data is already timesliced in the appropriate scheme, it will be returned
    unchanged.

    Args:
        data: Array to broadcast.
        ts: Dataarray with timeslice weights. If None, defaults to the global timeslice.
        level: Level to broadcast to. If None, use the finest level of ts.

    """
    from xarray import Coordinates

    if ts is None:
        ts = TIMESLICE

    if level is not None:
        ts = compress_timeslice(ts, ts=ts, level=level, operation="sum")

    # If data already has timeslices, check that it matches the reference timeslice.
    if "timeslice" in data.dims:
        if data.timeslice.reset_coords(drop=True).equals(
            ts.timeslice.reset_coords(drop=True)
        ):
            return data
        raise ValueError(
            "Data is already timesliced, but does not match the reference."
        )

    mindex_coords = Coordinates.from_pandas_multiindex(ts.timeslice, "timeslice")
    broadcasted = data.expand_dims(timeslice=ts["timeslice"]).assign_coords(
        mindex_coords
    )
    return broadcasted


def distribute_timeslice(
    data: DataArray, ts: DataArray | None = None, level=None
) -> DataArray:
    """Convert a non-timesliced array to a timesliced array by distribution.

    Takes non-timesliced data and distributes it over the timeslice dimension according
    to the timeslice weights in `ts`. The sum of the output over all timeslices will be
    equal to the input. If data is already timesliced in the appropriate scheme, it will
    be returned unchanged.

    Args:
        data: Array to distribute.
        ts: Dataarray with timeslice weights. If None, defaults to the global timeslice.
        level: Level to distribute to. If None, use the finest level of ts.

    """
    if ts is None:
        ts = TIMESLICE

    if level is not None:
        ts = compress_timeslice(ts, ts=ts, level=level, operation="sum")

    # If data already has timeslices, check that it matches the reference timeslice.
    if "timeslice" in data.dims:
        if data.timeslice.reset_coords(drop=True).equals(
            ts.timeslice.reset_coords(drop=True)
        ):
            return data
        raise ValueError(
            "Data is already timesliced, but does not match the reference."
        )

    broadcasted = broadcast_timeslice(data, ts=ts)
    timeslice_sum = ts.sum("timeslice").clip(1e-6)  # prevents zero division
    timeslice_fractions = ts / broadcast_timeslice(timeslice_sum, ts=ts)
    return broadcasted * timeslice_fractions


def compress_timeslice(
    data: DataArray,
    ts: DataArray | None = None,
    level: str | None = None,
    operation: str = "sum",
) -> DataArray:
    """Convert a fully timesliced array to a coarser level.

    The operation can be either 'sum', or 'mean':

    - sum: sum values at each compressed timeslice level
    - mean: take a weighted average of values at each compressed timeslice level,
        according to the timeslice weights in ts

    Args:
        data: Timesliced array to compress. Must have the same timeslicing as ts.
        ts: Dataarray with timeslice weights. If None, defaults to the global timeslice.
        level: Level to compress to. If None, don't compress.
        operation: Operation to perform ("sum" or "mean"). Defaults to "sum".

    """
    if ts is None:
        ts = TIMESLICE

    # Raise error if data is not timesliced appropriately
    if "timeslice" not in data.dims:
        raise ValueError("Data must have a 'timeslice' dimension.")
    if not data.timeslice.reset_coords(drop=True).equals(
        ts.timeslice.reset_coords(drop=True)
    ):
        raise ValueError("Data has incompatible timeslicing with reference.")

    # If level is not specified, don't compress
    if level is None:
        return data

    # level must be a valid timeslice level
    x_levels = data.timeslice.to_index().names
    if level not in x_levels:
        raise ValueError(f"Unknown level: {level}. Must be one of {x_levels}.")

    # Return data unchanged if already at the desired level
    if get_level(data) == level:
        return data

    # Prepare mask
    idx = x_levels.index(level)
    kept_levels, compressed_levels = x_levels[: idx + 1], x_levels[idx + 1 :]
    mask = ts.unstack(dim="timeslice")
    if operation == "sum":
        mask = mask.where(np.isnan(mask), 1)
    elif operation == "mean":
        mask = mask / mask.sum(compressed_levels)
    else:
        raise ValueError(f"Unknown operation: {operation}. Must be 'sum' or 'mean'.")

    # Perform the operation
    result = (
        (data.unstack(dim="timeslice") * mask)
        .sum(compressed_levels)
        .stack(timeslice=kept_levels)
    )
    return sort_timeslices(result, ts)


def expand_timeslice(
    data: DataArray, ts: DataArray | None = None, operation: str = "distribute"
) -> DataArray:
    """Convert a timesliced array to a finer level.

    The operation can be either 'distribute', or 'broadcast'

    - distribute: distribute values over the new timeslice level(s) according to
        timeslice weights in ``ts``, such that the sum of the output over all timeslices
        is equal to the sum of the input
    - broadcast: broadcast values across over the new timeslice level(s)

    Args:
        data: Timesliced array to expand.
        ts: Dataarray with timeslice weights. If None, defaults to the global timeslice.
        operation: Operation to perform ("distribute" or "broadcast").
            Defaults to "distribute".

    """
    if ts is None:
        ts = TIMESLICE

    # Raise error if data is not timesliced
    if "timeslice" not in data.dims:
        raise ValueError("Data must have a 'timeslice' dimension.")

    # Get level names
    ts_levels = ts.timeslice.to_index().names
    x_levels = data.timeslice.to_index().names

    # Raise error if x_levels is not a subset of ts_levels
    if not set(x_levels).issubset(ts_levels):
        raise ValueError(
            "Data has incompatible timeslicing with reference. "
            f"Timeslice levels of data ({x_levels}) must be a subset of ts "
            f"({ts_levels})."
        )

    # Return data unchanged if already at the desired level
    finest_level = get_level(ts)
    current_level = get_level(data)
    if current_level == finest_level:
        return data

    # Prepare mask
    mask = ts.unstack(dim="timeslice")
    if operation == "broadcast":
        mask = mask.where(np.isnan(mask), 1)
    elif operation == "distribute":
        mask = mask / mask.sum(ts_levels[ts_levels.index(current_level) + 1 :])
    else:
        raise ValueError(
            f"Unknown operation: {operation}. Must be 'distribute' or 'broadcast'."
        )

    # Perform the operation
    result = (
        (data.unstack(dim="timeslice") * mask)
        .stack(timeslice=ts_levels)
        .dropna("timeslice")
    )
    return sort_timeslices(result, ts)


def drop_timeslice(data: DataArray) -> DataArray:
    """Drop the timeslice variable from a DataArray.

    If the array doesn't contain the timeslice variable, return the input unchanged.
    """
    if "timeslice" not in data.dims:
        return data

    return data.drop_vars(data.timeslice.indexes)


def get_level(data: DataArray) -> str:
    """Get the timeslice level of a DataArray."""
    if "timeslice" not in data.dims:
        raise ValueError("Data does not have a 'timeslice' dimension.")
    return data.timeslice.to_index().names[-1]


def sort_timeslices(data: DataArray, ts: DataArray | None = None) -> DataArray:
    """Sorts the timeslices of a DataArray according to a reference timeslice.

    This will only sort timeslices to match the reference if the data is at the same
    timeslice level as the reference. Otherwise, it will sort timeslices in alphabetical
    order.

    Args:
        data: Timesliced DataArray to sort.
        ts: Dataarray with reference timeslices in the appropriate order
    """
    if ts is None:
        ts = TIMESLICE

    # If data is at the same timeslice level as ts, sort timeslices according to ts
    if get_level(data) == get_level(ts):
        return data.sel(timeslice=ts.timeslice)
    # Otherwise, sort timeslices in alphabetical order
    return data.sortby("timeslice")


def timeslice_max(data: DataArray, ts: DataArray | None = None) -> DataArray:
    """Find the max value over the timeslice dimension, normalized for timeslice length.

    This first annualizes the value in each timeslice by dividing by the fraction of the
    year that the timeslice occupies, then takes the maximum value

    Args:
        data: Timesliced DataArray to find the max of.
        ts: Dataarray with relative timeslice lengths. If None, defaults to the global
        timeslice.
    """
    if ts is None:
        ts = TIMESLICE

    timeslice_level = get_level(data)
    timeslice_sum = ts.sum("timeslice").clip(1e-6)  # prevents zero division
    timeslice_fractions = compress_timeslice(
        ts, ts=ts, level=timeslice_level, operation="sum"
    ) / broadcast_timeslice(timeslice_sum, ts=ts, level=timeslice_level)
    return (data / timeslice_fractions).max("timeslice")



"""Collection of functions and stand-alone algorithms."""

from __future__ import annotations

from collections.abc import Hashable, Iterable, Iterator, Mapping, Sequence
from typing import (
    Any,
    Callable,
    NamedTuple,
    cast,
)

import numpy as np
import xarray as xr


def multiindex_to_coords(data: xr.Dataset | xr.DataArray, dimension: str = "asset"):
    """Flattens multi-index dimension into multi-coord dimension."""
    from pandas import MultiIndex

    assert dimension in data.dims
    assert isinstance(data.indexes[dimension], MultiIndex)
    names = data.indexes[dimension].names
    coords = {n: data[n].values for n in names}
    result = data.drop_vars([dimension, *names])
    for name, coord in coords.items():
        result[name] = dimension, coord
    if isinstance(result, xr.Dataset):
        result = result.set_coords(names)
    return result


def coords_to_multiindex(
    data: xr.Dataset | xr.DataArray, dimension: str = "asset"
) -> xr.Dataset | xr.DataArray:
    """Creates a multi-index from flattened multiple coords."""
    from pandas import MultiIndex

    assert dimension in data.dims
    assert dimension not in data.indexes
    names = [u for u in data.coords if data[u].dims == (dimension,)]
    index = MultiIndex.from_arrays([data[u].values for u in names], names=names)
    mindex_coords = xr.Coordinates.from_pandas_multiindex(index, dimension)
    return data.drop_vars(names).assign_coords(mindex_coords)


def reduce_assets(
    assets: xr.DataArray | xr.Dataset | Sequence[xr.Dataset | xr.DataArray],
    coords: str | Sequence[str] | Iterable[str] | None = None,
    dim: str = "asset",
    operation: Callable | None = None,
) -> xr.DataArray | xr.Dataset:
    r"""Combine assets along given asset dimension.

    This method simplifies combining assets across multiple agents, or combining assets
    across a given dimension. By default, it will sum together assets from the same
    region which have the same technology and the same installation date. In other
    words, assets are identified by the technology, installation year and region. The
    reduction happens over other possible coordinates, e.g. the owning agent.

    More specifically, assets are often indexed using what xarray calls a **dimension
    without coordinates**. In practice, there are still coordinates associated with
    assets, e.g. *technology* and *installed* (installation year or version), but the
    value associated with these coordinates are not unique.  There may be more than one
    asset with the same technology or installation year.

    For instance, with assets per agent defined as :math:`A^{i, r}_o`, with :math:`i` an
    agent index, :math:`r` a region, :math:`o` is the coordinates identified in
    ``coords``, and :math:`T` the transformation identified by ``operation``, then this
    function computes:

    .. math::

        R_{r, o} = T[\{A^{i, r}_o; \forall i\}]

    If :math:`T` is the sum operation, then:

    .. math::

        R_{r, o} = \sum_i  A^{i, r}_o


    Example:
        Lets construct assets that do have duplicates assets. First we construct the
        dimensions, using fake data:

        >>> data = xr.Dataset()
        >>> data['year'] = 'year', [2010, 2015, 2017]
        >>> data['installed'] = 'asset', [1990, 1991, 1991, 1990]
        >>> data['technology'] = 'asset', ['a', 'b', 'b', 'c']
        >>> data['region'] = 'asset', ['x', 'x', 'x', 'y']
        >>> data = data.set_coords(('installed', 'technology', 'region'))

        We can check there are duplicate assets in this coordinate system:

        >>> processes = set(
        ...     zip(data.installed.values, data.technology.values, data.region.values)
        ... )
        >>> len(processes) < len(data.asset)
        True

        Now we can easily create a fake two dimensional quantity per process and
        per year:

        >>> data['capacity'] = ('year', 'asset'), np.arange(3 * 4).reshape(3, 4)

        The point of `reduce_assets` is to aggregate assets that refer to the
        same process:

        >>> reduce_assets(data.capacity)  # doctest: +SKIP
        <xarray.DataArray 'capacity' (year: 3, asset: 3)> Size: 36B
        array([[ 0,  3,  3],
               [ 4,  7, 11],
               [ 8, 11, 19]])
        Coordinates:
          * year        (year) int32 12B 2010 2015 2017
            installed   (asset) int32 12B 1990 1990 1991
            technology  (asset) <U1 12B 'a' 'c' 'b'
            region      (asset) <U1 12B 'x' 'y' 'x'
        Dimensions without coordinates: asset

        We can also specify explicitly which coordinates in the 'asset'
        dimension should be reduced, and how:

        >>> reduce_assets(
        ...     data.capacity,
        ...     coords=('technology', 'installed'),
        ...     operation = lambda x: x.mean(dim='asset')
        ... )  # doctest: +SKIP
        <xarray.DataArray 'capacity' (year: 3, asset: 3)> Size: 72B
        array([[ 0. ,  1.5,  3. ],
               [ 4. ,  5.5,  7. ],
               [ 8. ,  9.5, 11. ]])
        Coordinates:
          * year        (year) int32 12B 2010 2015 2017
            technology  (asset) <U1 12B 'a' 'b' 'c'
            installed   (asset) int32 12B 1990 1991 1990
        Dimensions without coordinates: asset
    """
    from copy import deepcopy

    assets = deepcopy(assets)

    if operation is None:

        def operation(x):
            return x.sum(dim)

    assert operation is not None

    # Concatenate assets if a sequence is given
    if not isinstance(assets, (xr.Dataset, xr.DataArray)):
        assets = xr.concat(assets, dim=dim)
    assert isinstance(assets, (xr.Dataset, xr.DataArray))

    # If there are no assets, nothing needs to be done
    if assets[dim].size == 0:
        return assets

    # Coordinates to reduce over (e.g. technology, installed)
    if coords is None:
        coords = [cast(str, k) for k, v in assets.coords.items() if v.dims == (dim,)]
    elif isinstance(coords, str):
        coords = (coords,)
    coords = [k for k in coords if k in assets.coords and assets[k].dims == (dim,)]

    # Create a new dimension to group by
    dtypes = [(d, assets[d].dtype) for d in coords]
    grouper = np.array(
        list(zip(*(cast(Iterator, assets[d].values) for d in coords))), dtype=dtypes
    )
    assert "grouper" not in assets.coords
    assets["grouper"] = "asset", grouper

    # Perform the operation
    result = operation(assets.groupby("grouper")).rename(grouper=dim)
    for i, d in enumerate(coords):
        result[d] = dim, [u[i] for u in result[dim].values]
    return result.drop_vars("asset")


def broadcast_over_assets(
    data: xr.Dataset | xr.DataArray,
    template: xr.DataArray | xr.Dataset,
    installed_as_year: bool = True,
) -> xr.Dataset | xr.DataArray:
    """Broadcasts an array to the shape of a template containing asset-level data.

    The dimensions of many arrays (such as technology datasets) are fully explicit, in
    that each concept (e.g. 'technology', 'region', 'year') is a separate dimension.
    However, other datasets (e.g capacity), are presented on a per-asset basis,
    containing a single 'asset' dimension with with coordinates such as 'region',
    'installed' (year of installation), and 'technology'. This latter representation is
    sparse if not all combinations of 'region', 'installed' and 'technology' are
    present.

    This function broadcasts the first representation to the shape and coordinates
    of the second, selecting the appropriate values for each asset (see example below).

    Note: this is not necessarily limited to technology datasets. For
    example, it could also be used on a dataset of commodity prices to select prices
    relevant to each asset (e.g. if assets exist in multiple regions).

    Args:
        data: The dataset/data-array to broadcast.
        template: The dataset/data-array to use as a template.
        installed_as_year: True means that the ``year`` dimension in ``data``
            corresponds to the year that the asset was installed. This will commonly
            be the case for most technology parameters (e.g. ``var_par``/``fix_par`` are
            specified for the year that an asset is installed, and fixed for the
            lifetime of the asset). In this case, ``data`` must have a ``year``
            coordinate for every possible ``installed`` year in the template.
            Conversely, if the values in ``data`` apply to the year of activity, rather
            than the year of installation, ``installed_as_year`` should be False.
            An example would be commodity prices, which can change over the lifetime
            of an asset. In this case, if ``year`` is present as a dimension in
            ``data``, it will be maintained as a separate dimension in the output.

    Example:
        Define the data array:
        >>> import xarray as xr
        >>> technologies = xr.DataArray(
        ...     data=[[1, 2, 3], [4, 5, 6]],
        ...     dims=['technology', 'region'],
        ...     coords={'technology': ['gasboiler', 'heatpump'],
        ...             'region': ['R1', 'R2', 'R3']},
        ... )

        This array contains a value for every combination of technology and region (e.g.
        this could refer to the efficiency of each technology in each region). For
        simplicity, we are not including a "year" dimension in this example.

        Define the assets template:
        >>> assets = xr.DataArray(
        ...     data=[10, 50],
        ...     dims=["asset"],
        ...     coords={
        ...         "region": (["asset"], ["R1", "R2"]),
        ...         "technology": (["asset"], ["gasboiler", "heatpump"])},
        ... )

        We have two assets: a gas boiler in region R1 and a heat pump in region R2. In
        this case the values don't matter, but could correspond to the installed
        capacity of each asset, for example.

        We want to select the values from the technology array that correspond to each
        asset in the template. To do this, we perform `broadcast_over_assets` on
        `technologies` using `assets` as a template:
        >>> broadcast_over_assets(technologies, assets, installed_as_year=False)
        <xarray.DataArray (asset: 2)> Size: 16B
        array([1, 5])
        Coordinates:
            technology  (asset) <U9 72B 'gasboiler' 'heatpump'
            region      (asset) <U2 16B 'R1' 'R2'
        Dimensions without coordinates: asset

        The output array has an "asset" dimension which matches the template. Each value
        in the output is the value in the original technology array that matches the
        technology & region of each asset.
    """
    # TODO: this will return `data` unchanged if the template has no "asset"
    # dimension, but strictly speaking we shouldn't allow this.
    # assert "asset" in template.dims

    # Name of asset coordinates (e.g. "technology", "region", "installed")
    names = [u for u in template.coords if template[u].dims == ("asset",)]
    assert "year" not in names

    # If installed_as_year is True, we need to rename the installed dimension to "year"
    # TODO: this should be stricter, and enforce that the template has "installed" data,
    # and that the technologies dataset has a "year" dimension.
    # if installed_as_year:
    if installed_as_year and "installed" in names and "year" in data.dims:
        # assert "installed" in names
        data = data.rename(year="installed")

    # The first selection reduces the size of the data without affecting the
    # dimensions.
    first_sel = {n: data[n].isin(template[n]) for n in names if n in data.dims}
    techs = data.sel(first_sel)

    # Reshape the array to match the template
    second_sel = {n: template[n] for n in template.coords if n in techs.dims}
    return techs.sel(second_sel)


def clean_assets(assets: xr.Dataset, year: int):
    """Cleans up and prepares asset for current iteration.

    - removes data from before the specified year
    - removes assets for which there is no capacity now or in the future
    """
    assets = assets.sel(year=slice(year, None))
    assets = assets.where(assets.capacity.any(dim="year"), drop=True)
    return assets


def tupled_dimension(array: np.ndarray, axis: int):
    """Transforms one axis into a tuples."""
    if array.shape[axis] == 1:
        shape = tuple(j for i, j in enumerate(array.shape) if i != axis)
        return array.reshape(*shape)

    rolled = np.moveaxis(array, axis, -1)
    shape = rolled.shape
    flattened = rolled.reshape(-1, shape[-1])
    result = np.zeros(shape=flattened.shape[:-1], dtype=object)
    for i in range(0, flattened.shape[0]):
        result[i] = tuple(flattened[i, :])
    return result.reshape(*shape[:-1])


def lexical_comparison(
    objectives: xr.Dataset,
    binsize: xr.Dataset,
    order: Sequence[Hashable] | None = None,
    bin_last: bool = True,
) -> xr.DataArray:
    """Lexical comparison over the objectives.

    Lexical comparison operates by binning the objectives into bins of width
    `binsize`. Once binned, dimensions other than `asset` and `technology` are
    reduced by taking the max, e.g. the largest constraint. Finally, the
    objectives are ranked lexographically, in the order given by the parameters.

    Arguments:
        objectives: xr.Dataset containing the objectives to rank
        binsize: bin size, minimization direction
            (+ -> minimize, - -> maximize), and (optionally) order of
            lexicographical comparison. The order is the one given
            `binsize.data_vars` if the argument `order` is None.
        order: Optional array indicating the order in which to rank the tuples.
        bin_last: Whether the last metric should be binned, or whether it
            should be left as a the type it already is (e.g. no flooring and
            no turning to integer.)

    Result:
        An array of tuples which can subsequently be compared lexicographically.
    """
    if order is None:
        order = [u for u in binsize.data_vars]

    assert set(order) == set(binsize.data_vars)
    assert set(order).issuperset(objectives)

    result = objectives[order]
    for name in order if bin_last else order[:-1]:
        result[name] = np.floor(result[name] / binsize[name]).astype(int)
    if not bin_last:
        result[order[-1]] = result[order[-1]] / binsize[order[-1]]
    return result.to_array(dim="variable").reduce(tupled_dimension, dim="variable")


def merge_assets(
    capa_a: xr.DataArray,
    capa_b: xr.DataArray,
    dimension: str = "asset",
) -> xr.DataArray:
    """Merge two capacity arrays."""
    # Interpolate capacity arrays to a common time framework
    years = sorted(set(capa_a.year.values).union(capa_b.year.values))
    if len(capa_a.year) == 1:
        capa_a_interp = capa_a
        capa_b_interp = interpolate_capacity(capa_b, year=years)
    elif len(capa_b.year) == 1:
        capa_a_interp = interpolate_capacity(capa_a, year=years)
        capa_b_interp = capa_b
    else:
        capa_a_interp = interpolate_capacity(capa_a, year=years)
        capa_b_interp = interpolate_capacity(capa_b, year=years)

    # Concatenate the two capacity arrays
    result = xr.concat((capa_a_interp, capa_b_interp), dim=dimension)

    #
    forgroup = result.pipe(coords_to_multiindex, dimension=dimension)
    if isinstance(forgroup, xr.DataArray):
        forgroup = forgroup.to_dataset()
    if len(forgroup[dimension]) != len(set(forgroup[dimension].values)):
        result = (
            forgroup.groupby(dimension)
            .sum(dimension)
            .clip(min=0)
            .pipe(multiindex_to_coords, dimension=dimension)
        )
    return result


def avoid_repetitions(data: xr.DataArray, dim: str = "year") -> xr.DataArray:
    """List of years such that there is no repetition in the data.

    It removes the central year of any three consecutive years where all data is
    the same. This means the original data can be reobtained via a linear
    interpolation or a forward fill.
    See :py:func:`muse.utilities.interpolate_capacity`.

    The first and last year are always preserved.
    """
    roll = data.rolling({dim: 3}, center=True).construct("window")
    years = ~(roll == roll.isel(window=0)).all([u for u in roll.dims if u != dim])
    return data.year[years]


def interpolate_capacity(
    data: xr.DataArray, year: int | Sequence[int] | xr.DataArray
) -> xr.DataArray:
    """Interpolates capacity data to the given years.

    Capacity between years is interpolated linearly. Capacity outside the range of the
    data is set to zero.

    Arguments:
        data: DataArray containing the capacity data
        year: Year or years to interpolate to
    """
    return data.interp(
        year=year,
        method="linear",
        kwargs={"fill_value": 0.0},
    )


def nametuple_to_dict(nametup: Mapping | NamedTuple) -> Mapping:
    """Transforms a nametuple of type GenericDict into an OrderDict."""
    from collections import OrderedDict
    from dataclasses import asdict, is_dataclass

    if is_dataclass(nametup):
        out = asdict(nametup, OrderedDict)  # type: ignore
    elif hasattr(nametup, "_asdict"):
        out = nametup._asdict()  # type: ignore
    else:
        out = nametup.copy()  # type: ignore
    for key, value in out.items():
        if is_dataclass(value) or hasattr(value, "_asdict"):
            out[key] = nametuple_to_dict(value)
    return out


def future_propagation(
    data: xr.DataArray,
    future: xr.DataArray,
    threshold: float = 1e-12,
) -> xr.DataArray:
    """Propagates values into the future.

    Example:
        ``Data`` should be an array with at least one dimension, "year":

        >>> coords = dict(year=list(range(2020, 2040, 5)), fuel=["gas", "coal"])
        >>> data = xr.DataArray(
        ...     [list(range(4)), list(range(-5, -1))],
        ...     coords=coords,
        ...     dims=("fuel", "year")
        ... )

        ``future`` is an array with  exactly one year in its ``year`` coordinate, or
        that coordinate must correspond to a scalar. That one year should also be
        present in ``data``.

        >>> future = xr.DataArray(
        ...     [1.2, -3.95], coords=dict(fuel=coords['fuel'], year=2025), dims="fuel",
        ... )

        This function propagates into ``data`` values from ``future``, but only if those
        values differed for the current year beyond a given threshold:

        >>> from muse.utilities import future_propagation
        >>> future_propagation(data, future, threshold=0.1)  # doctest: +SKIP
        <xarray.DataArray (fuel: 2, year: 4)> Size: 64B
        array([[ 0. ,  1.2,  1.2,  1.2],
               [-5. , -4. , -3. , -2. ]])
        Coordinates:
          * year     (year) int32 16B 2020 2025 2030 2035
          * fuel     (fuel) <U4 32B 'gas' 'coal'

        Above, the data for coal is not sufficiently different given the threshold.
        hence, the future values for coal remain as they where.

        The dimensions of ``future`` do not have to match exactly those of ``data``.
        Standard broadcasting is used if they do not match:

        >>> future_propagation(
        ...    data, future.sel(fuel="gas", drop=True), threshold=0.1
        ... )  # doctest: +SKIP
        <xarray.DataArray (fuel: 2, year: 4)> Size: 64B
        array([[ 0. ,  1.2,  1.2,  1.2],
               [-5. ,  1.2,  1.2,  1.2]])
        Coordinates:
          * year     (year) int32 16B 2020 2025 2030 2035
          * fuel     (fuel) <U4 32B 'gas' 'coal'
        >>> future_propagation(
        ...     data, future.sel(fuel="coal", drop=True), threshold=0.1
        ... )  # doctest: +SKIP
        <xarray.DataArray (fuel: 2, year: 4)> Size: 64B
        array([[ 0.  , -3.95, -3.95, -3.95],
               [-5.  , -4.  , -3.  , -2.  ]])
        Coordinates:
          * year     (year) int32 16B 2020 2025 2030 2035
          * fuel     (fuel) <U4 32B 'gas' 'coal'
    """
    if "year" not in data.dims or "year" not in future.coords:
        raise ValueError("Expected dimension 'year' in `data` and `future`.")
    if future["year"].ndim != 0 and len(future["year"]) != 1:
        raise ValueError('``future["year"] should be of length 1 or a scalar.')
    if not future["year"].isin(data["year"]).all():
        raise ValueError(f'{future["year"]} not found in data["year"].')

    year = future["year"].item()
    if "year" in future.dims:
        future = future.sel(year=year, drop=True)
    return data.where(
        np.logical_or(
            data.year < year, np.abs(data.sel(year=year) - future) < threshold
        ),
        future,
    )


def agent_concatenation(
    data: Mapping[Hashable, xr.DataArray | xr.Dataset],
    dim: str = "asset",
    name: str = "agent",
    fill_value: Any = 0,
) -> xr.DataArray | xr.Dataset:
    """Concatenates input map along given dimension.

    Example:
        Lets create sets of random assets to work with. We set the seed so that this
        test can be reproduced exactly.

        >>> from muse.examples import random_agent_assets
        >>> rng = np.random.default_rng(1234)
        >>> assets = {i: random_agent_assets(rng) for i in range(5)}

        The concatenation will create a new dataset (or datarray) combining all the
        inputs along the dimension "asset". The origin of each datum is retained in a
        new coordinate "agent" with dimension "asset".

        >>> from muse.utilities import agent_concatenation
        >>> aggregate = agent_concatenation(assets)
        >>> aggregate # doctest: +SKIP
        <xarray.Dataset> Size: 4kB
        Dimensions:     (asset: 19, year: 12)
        Coordinates:
            agent       (asset) int32 76B 0 0 0 0 0 1 1 1 2 2 2 2 3 3 3 4 4 4 4
          * year        (year) int64 96B 2033 2035 2036 2037 ... 2046 2047 2048 2049
            installed   (asset) int64 152B 2030 2025 2030 2010 ... 2025 2030 2010 2025
            technology  (asset) <U9 684B 'oven' 'stove' 'oven' ... 'oven' 'thermomix'
            region      (asset) <U9 684B 'Brexitham' 'Brexitham' ... 'Brexitham'
        Dimensions without coordinates: asset
        Data variables:
            capacity    (asset, year) float64 2kB 26.0 26.0 26.0 56.0 ... 62.0 62.0 62.0

        Note that the `dtype` of the capacity has changed from integers to floating
        points. This is due to how ``xarray`` performs the operation.

        We can check that all the data from each agent is indeed present in the
        aggregate.

        >>> for agent, inventory in assets.items():
        ...    assert (aggregate.sel(asset=aggregate.agent == agent) == inventory).all()

        However, it should be noted that the data is not always strictly equivalent:
        dimensions outside of "assets" (most notably "year") will include all points
        from all agents. Missing values for the "year" dimension are forward filled (and
        backfilled with zeros). Others are left with "NaN".
    """
    from itertools import repeat

    data = {k: v.copy() for k, v in data.items()}
    for key, datum in data.items():
        if name in datum.coords:
            raise ValueError(f"Coordinate {name} already exists")
        if len(data) == 1 and isinstance(datum, xr.DataArray):
            data[key] = datum.assign_coords(
                {
                    name: (
                        dim,
                        list(repeat(key, datum.sizes[dim])),
                    )
                }
            )
        else:
            datum[name] = key
    result = xr.concat(data.values(), dim=dim)
    if isinstance(result, xr.Dataset):
        result = result.set_coords("agent")
    if "year" in result.dims:
        result = result.ffill("year")
    if fill_value is not np.nan:
        result = result.fillna(fill_value)
    return result


def check_dimensions(
    data: xr.DataArray | xr.Dataset,
    required: Iterable[str] = (),
    optional: Iterable[str] = (),
):
    """Ensure that an array has the required dimensions.

    This will check that all required dimensions are present, and that no other
    dimensions are present, apart from those listed as optional.

    Args:
        data: DataArray or Dataset to check dimensions of
        required: List of dimension names that must be present
        optional: List of dimension names that may be present
    """
    present = set(data.dims)
    missing = set(required) - present
    if missing:
        raise ValueError(f"Missing required dimensions: {missing}")
    extra = present - set(required) - set(optional)
    if extra:
        raise ValueError(f"Extra dimensions: {extra}")


def interpolate_technodata(
    data: xr.Dataset,
    time_framework: list[int],
    interpolation_mode: str = "linear",
) -> xr.Dataset:
    """Interpolates technologies data to a given time framework.

    The approach depends on the format of the data:
    - If the original data contains data for more than one year, then data will be
    interpolated with flat back/forward extension to cover the time period.
    - If the original data does not have a "year" dimension, or only has data for a
    single year, then this data will not be modified or duplicated, but a dummy
    "year" dimension covering the time framework will be added.

    In both cases, data for any year in the time framework can be selected
    using data.sel(year=year) on the returned dataset.

    Args:
        data: Dataset to interpolate
        time_framework: List of years to interpolate to
        interpolation_mode: Interpolation mode to use. Must be one of:
            "linear", "nearest", "zero", "slinear", "quadratic", "cubic"

    Returns:
        Dataset with the data interpolated to the time framework, sorted by year.
    """
    # If the data has only one year then no interpolation is needed,  we just need to
    # add a year dimension with the time framework
    if "year" not in data.dims:
        data = data.copy()
        data["year"] = ("year", time_framework)
        return data

    if len(data.year) == 1:
        data = data.isel(year=0, drop=True)
        data["year"] = ("year", time_framework)
        return data

    # Sort the data by year, just in case it hasn't already been sorted
    data = data.sortby("year")

    # Flat forward extrapolation
    maxyear = max(time_framework)
    if data.year.max() < maxyear:
        years = [*data.year.values.tolist(), maxyear]
        data = data.reindex(year=years, method="ffill")

    # Flat backward extrapolation
    minyear = min(time_framework)
    if data.year.min() > minyear:
        years = [minyear, *data.year.values.tolist()]
        data = data.reindex(year=years, method="bfill")

    # Interpolation to fill gaps
    years = sorted(set(time_framework).union(data.year.values.tolist()))
    data = data.interp(year=years, method=interpolation_mode)
    return data


def camel_to_snake(name: str) -> str:
    """Transforms CamelCase to snake_case."""
    from re import sub

    pattern = sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    result = sub("([a-z0-9])([A-Z])", r"\1_\2", pattern).lower()
    result = result.replace("co2", "CO2")
    result = result.replace("ch4", "CH4")
    result = result.replace("n2_o", "N2O")
    result = result.replace("f-gases", "F-gases")
    return result



"""Default global values used in keyword arguments."""

from pathlib import Path

DEFAULT_SECTORS_DIRECTORY = Path().cwd() / "data"

DATA_DIRECTORY = Path(__file__).parent / "data"
""" Standard data directory."""

""" Default root directory with sector data """
DEFAULT_OUTPUT_DIRECTORY = Path("Results")
""" Default root directory with sector data """
DEFAULT_SETTINGS_DIRECTORY = Path.cwd()
""" Directory where to look for `settings.toml`."""



"""Demand share computations.

.. currentmodule:: muse.demand_share

The demand share splits a demand amongst agents. It is used within a sector to assign
part of the input MCA demand to each agent.

Demand shares functions should be registered via the decorator `register_demand_share`.

Demand share functions are not expected to modify any of their arguments. They
should all have the following signature:

.. code-block:: Python

    @register_demand_share
    def demand_share(
        agents: Sequence[AbstractAgent],
        demand: xr.Dataarray,
        technologies: xr.Dataset,
        **kwargs
    ) -> xr.DataArray:
        pass

Arguments:
    agents: a sequence of  agent relevant to the demand share procedure. The agent can
        be queried for parameters specific to the demand share procedure. For instance,
        :py:func`new_and_retro` will query the agents for the assets they own, the
        region they are contained with, their category (new or retrofit), etc...
    demand: DataArray of commodity demands for the current year and investment year.
    technologies: a dataset containing all constant data characterizing the
        technologies.
    kwargs: Additional keyword arguments.

Returns:
    A DataArray of demand shares.
"""

from __future__ import annotations

from collections.abc import Hashable, MutableMapping, Sequence
from functools import wraps
from logging import getLogger
from typing import Any, Callable, cast

import numpy as np
import xarray as xr
from mypy_extensions import KwArg

from muse.agents import AbstractAgent
from muse.commodities import is_enduse
from muse.errors import RetrofitAgentInStandardDemandShare
from muse.quantities import maximum_production
from muse.registration import registrator
from muse.utilities import (
    agent_concatenation,
    broadcast_over_assets,
    check_dimensions,
    interpolate_capacity,
    reduce_assets,
)

__all__ = [
    "DEMAND_SHARE_SIGNATURE",
    "factory",
    "new_and_retro",
    "register_demand_share",
    "standard_demand",
    "unmet_demand",
    "unmet_forecasted_demand",
]

DEMAND_SHARE_SIGNATURE = Callable[
    [Sequence[AbstractAgent], xr.DataArray, xr.Dataset, KwArg(Any)], xr.DataArray
]
"""Demand share signature."""

DEMAND_SHARE: MutableMapping[str, DEMAND_SHARE_SIGNATURE] = {}
"""Dictionary of demand share functions."""


@registrator(registry=DEMAND_SHARE, loglevel="info")
def register_demand_share(function: DEMAND_SHARE_SIGNATURE) -> DEMAND_SHARE_SIGNATURE:
    """Registers a demand share function with MUSE."""

    @wraps(function)
    def decorated(
        agents: Sequence[AbstractAgent],
        demand: xr.DataArray,
        technologies: xr.Dataset,
        **kwargs,
    ) -> xr.DataArray:
        """Computes and validates a demand share."""
        # Validate inputs
        check_dimensions(
            demand,
            ["commodity", "year", "timeslice", "region"],
            optional=["dst_region"],
        )
        assert len(demand.year) == 2
        check_dimensions(
            technologies,
            ["technology", "region"],
            optional=["timeslice", "commodity", "dst_region"],
        )

        # We can only share demand for enduse commodities
        # So we need to check that demand is zero for all non-enduse commodities
        enduse_names = technologies.commodity[is_enduse(technologies.comm_usage)]
        non_enduse = demand.commodity[~demand.commodity.isin(enduse_names)]
        assert (demand.sel(commodity=non_enduse) == 0).all()

        # Calculate demand share
        result = function(agents, demand, technologies, **kwargs)

        # Validate output
        check_dimensions(
            result, ["timeslice", "commodity"], optional=["asset", "region"]
        )  # TODO: asset should be required, but trade model is failing
        return result

    return cast(DEMAND_SHARE_SIGNATURE, decorated)


def factory(name: str) -> DEMAND_SHARE_SIGNATURE:
    """Get a demand share function by name."""
    return DEMAND_SHARE[name]


@register_demand_share(name="new_and_retro")
def new_and_retro(
    agents: Sequence[AbstractAgent],
    demand: xr.DataArray,
    technologies: xr.Dataset,
    timeslice_level: str | None = None,
) -> xr.DataArray:
    r"""Splits demand across new and retro agents.

    The input demand is split amongst both *new* and *retrofit* agents. *New* agents get
    a share of the increase in demand for the investment year, whereas *retrofit* agents
    are assigned a share of the demand that occurs from decommissioned assets.

    Args:
        agents: a list of all agents. This list should mainly be used to determine the
            type of an agent and the assets it owns. The agents will not be modified in
            any way.
        demand: commodity demands for the current year and investment year.
        technologies: quantities describing the technologies.
        timeslice_level: the timeslice level of the sector (e.g. "hour", "day")

    Pseudo-code:

    #. the capacity is reduced over agents and  expanded over timeslices (extensive
       quantity) and aggregated over agents. Generally:

       .. math::

           A_{a, s}^r = w_s\sum_i A_a^{r, i}

       with :math:`w_s` a weight associated with each timeslice and determined via
       :py:func:`muse.timeslices.distribute_timeslice`.

    #. An intermediate quantity, the :py:func:`unmet demand
       <muse.demand_share.unmet_demand>` :math:`U` is defined from
       :math:`P[\mathcal{M}, \mathcal{A}]`, a function giving the production for a given
       market :math:`\mathcal{M}`, the associated consumption :math:`\mathcal{C}`, and
       aggregate assets :math:`\mathcal{A}`:

       .. math::
           U[\mathcal{M}, \mathcal{A}] =
             \max(\mathcal{C} - P[\mathcal{M}, \mathcal{A}], 0)

       where :math:`\max` operates element-wise, and indices have been dropped for
       simplicity. The resulting expression has the same indices as the consumption
       :math:`\mathcal{C}_{c, s}^r`.

       :math:`P` is the maximum production, given by
       <muse.quantities.maximum_production>`.


    #. the *new* demand :math:`N` is defined as:

        .. math::

            N = \min\left(
                \mathcal{C}_{c, s}^r(y + \Delta y) - \mathcal{C}_{c, s}^r(y),
                U[\mathcal{M}^r(y + \Delta y), \mathcal{A}_{a, s}^r(y)]
            \right)

    #. the *retrofit* demand :math:`R` is defined from the identity

       .. math::

           C_{c, s}^r(y + \Delta y) =
            P[\mathcal{M}^r(y+\Delta y), \mathcal{A}_{a, s}^r(y + \Delta y)]
            + N_{c, s}^r
            + R_{c, s}^r

       In other words, it is the share of the forecasted consumption that is serviced
       neither by the current assets still present in the investment year, nor by the
       *new* agent.

    #. then each *new* agent gets a share of :math:`N` proportional to it's
        share of the :py:func:`production <muse.quantities.maximum_production>`,
        :math:`P[\mathcal{A}_{a, s}^{r, i}(y)]`.  Then the share of the demand for new
        agent :math:`i` is:

        .. math::

            N_{c, s, t}^{i, r}(y) = N_{c, s}^r
                \frac{\sum_\iota P[\mathcal{A}_{s, t, \iota}^{r, i}(y)]}
                     {\sum_{i, t, \iota}P[\mathcal{A}_{s, t, \iota}^{r, i}(y)]}


    #. similarly, each *retrofit* agent gets a share of :math:`N` proportional to its
        share of the ``decommissioning_demand``, :math:`D^{r, i}_{t, c}`.
        Then the share of the demand for retrofit agent :math:`i` is:

        .. math::

            R_{c, s, t}^{i, r}(y) = R_{c, s}^r
                \frac{\sum_\iota\mathcal{D}_{t, c, \iota}^{i, r}(y)}
                    {\sum_{i, t, \iota}\mathcal{D}_{t, c, \iota}^{i, r}(y)}


    Note that in the last two steps, the assets owned by the agent are aggregated over
    the installation year. The effect is that the demand serviced by agents is
    disaggregated over each technology, rather than not over each *model* of each
    technology (asset).

    .. seealso::

        ``decommissioning_demand``,
        :py:func:`muse.quantities.maximum_production`
    """
    current_year, investment_year = map(int, demand.year.values)

    # Calculate existing capacity and broadcast technologies
    capacity = interpolate_capacity(
        reduce_assets([agent.assets.capacity for agent in agents]),
        year=[current_year, investment_year],
    )
    technodata = broadcast_over_assets(technologies, capacity, installed_as_year=True)

    # Calculate new and retrofit demands
    demands = new_and_retro_demands(
        capacity=capacity,
        demand=demand,
        technologies=technodata,
        timeslice_level=timeslice_level,
    )

    # Split demand between agents
    agent_demands: MutableMapping[Hashable, xr.DataArray] = {}
    for region in demands.region.values:
        total_retro_quantity = 0
        total_new_quantity = 0
        for agent in agents:
            if agent.region != region:
                continue

            # Select data for the agent
            current_capacity = interpolate_capacity(
                agent.assets.capacity, year=[current_year, investment_year]
            )
            current_technodata = technodata.sel(asset=agent.assets.asset)

            # Calculate the agent's share of the retrofit and new demands
            agent_retrofit_demand = demands.retrofit.sel(region=region) * agent.quantity
            agent_new_demand = demands.new.sel(region=region) * agent.quantity

            if agent.category == "retrofit":
                total_retro_quantity += agent.quantity
                # Split retrofit demand over assets based on decommissioning demand
                retro_shares = decommissioning_demand(
                    technologies=current_technodata,
                    capacity=interpolate_capacity(
                        current_capacity, year=[current_year, investment_year]
                    ),
                    timeslice_level=timeslice_level,
                )
                agent_demands[agent.uuid] = _inner_split(
                    agent_retrofit_demand,
                    retro_shares,
                )
            elif agent.category == "newcapa":
                total_new_quantity += agent.quantity
                # Split new demand over assets based on maximum production
                new_shares = maximum_production(
                    capacity=current_capacity,
                    technologies=current_technodata,
                    year=current_year,
                    timeslice_level=timeslice_level,
                )
                agent_demands[agent.uuid] = _inner_split(
                    agent_new_demand,
                    new_shares,
                )
            else:
                raise ValueError(f"Unknown agent category: {agent.category}")

        # Make sure the total new/retro agent quantity = 1
        # TODO: ideally we should check this in the input layer
        if abs(total_retro_quantity - 1) > 1e-2:
            msg = (
                f"Total retrofit agent quantity in region {region} "
                f"is not 1: {total_retro_quantity}"
            )
            getLogger(__name__).critical(msg)
        if abs(total_new_quantity - 1) > 1e-2:
            msg = (
                f"Total new agent quantity in region {region} "
                f"is not 1: {total_new_quantity}"
            )
            getLogger(__name__).critical(msg)

    result = agent_concatenation(agent_demands)
    assert "year" not in result.dims
    return result


@register_demand_share(name="default")
def standard_demand(
    agents: Sequence[AbstractAgent],
    demand: xr.DataArray,
    technologies: xr.Dataset,
    timeslice_level: str | None = None,
) -> xr.DataArray:
    r"""Splits demand across new agents.

    The input demand is split amongst *new* agents. *New* agents get a
    share of the increase in demand for the investment year, as well as the demand that
    occurs from decommissioned assets.

    Args:
        agents: a list of all agents. This list should mainly be used to determine the
            type of an agent and the assets it owns. The agents will not be modified in
            any way.
        demand: commodity demands for the current year and investment year.
        technologies: quantities describing the technologies.
        timeslice_level: the timeslice level of the sector (e.g. "hour", "day")

    """
    # Validate no retrofit agents
    if any(agent.category == "retrofit" for agent in agents):
        raise RetrofitAgentInStandardDemandShare(
            "Standard demand share cannot be used with retrofit agents"
        )

    current_year, investment_year = map(int, demand.year.values)

    # Calculate existing capacity and broadcast technologies
    capacity = interpolate_capacity(
        reduce_assets([agent.assets.capacity for agent in agents]),
        year=[current_year, investment_year],
    )
    technodata = broadcast_over_assets(technologies, capacity, installed_as_year=True)

    # Calculate new and retrofit demands
    demands = new_and_retro_demands(
        capacity=capacity,
        demand=demand,
        technologies=technodata,
        timeslice_level=timeslice_level,
    )

    # Split demand between agents
    agent_demands: MutableMapping[Hashable, xr.DataArray] = {}
    for region in demands.region.values:
        total_quantity = 0
        for agent in agents:
            if agent.region != region:
                continue

            # Select data for the agent
            agent_capacity = interpolate_capacity(
                agent.assets.capacity, year=[current_year, investment_year]
            )
            agent_technodata = technodata.sel(asset=agent.assets.asset)

            # Calculate the agent's share of the retrofit and new demands
            agent_retrofit_demand = demands.retrofit.sel(region=region) * agent.quantity
            agent_new_demand = demands.new.sel(region=region) * agent.quantity
            total_quantity += agent.quantity

            # Split retrofit demand over assets based on decommissioning demand
            retro_shares = decommissioning_demand(
                technologies=agent_technodata,
                capacity=agent_capacity,
                timeslice_level=timeslice_level,
            )
            retro_demands = _inner_split(
                agent_retrofit_demand,
                retro_shares,
            )

            # Split new demand over assets based on maximum production
            new_shares = maximum_production(
                capacity=agent_capacity,
                technologies=agent_technodata,
                year=current_year,
                timeslice_level=timeslice_level,
            )
            new_demands = _inner_split(
                agent_new_demand,
                new_shares,
            )

            # Sum new and retrofit demands for the agent
            agent_demands[agent.uuid] = retro_demands + new_demands

        # Make sure the total agent quantity = 1
        # TODO: ideally we should check this in the input layer
        if abs(total_quantity - 1) > 1e-2:
            msg = f"Total agent quantity in region {region} is not 1: {total_quantity}"
            getLogger(__name__).critical(msg)

    result = agent_concatenation(agent_demands)
    assert "year" not in result.dims
    return result


@register_demand_share(name="unmet_demand")
def unmet_forecasted_demand(
    agents: Sequence[AbstractAgent],
    demand: xr.DataArray,
    technologies: xr.Dataset,
    timeslice_level: str | None = None,
) -> xr.DataArray:
    """Forecast demand that cannot be serviced by non-decommissioned current assets."""
    current_year, investment_year = map(int, demand.year.values)

    # Calculate existing capacity
    capacity = interpolate_capacity(
        reduce_assets([agent.assets.capacity for agent in agents]),
        year=[current_year, investment_year],
    )

    # Select data for future years
    future_demand = demand.sel(year=investment_year, drop=True)
    future_capacity = capacity.sel(year=investment_year)

    # Calculate unmet demand
    result = unmet_demand(
        demand=future_demand,
        capacity=future_capacity,
        technologies=broadcast_over_assets(
            technologies, capacity, installed_as_year=True
        ),
        timeslice_level=timeslice_level,
    )
    assert "year" not in result.dims
    return result


def _inner_split(
    demand: xr.DataArray,
    shares: xr.DataArray,
) -> xr.DataArray:
    """Split demand over assets based on shares.

    Args:
        demand: the demand to split
        shares: the shares to apply to each asset
    """
    # Assets of the same technology type are grouped together
    grouped_shares: xr.DataArray = (
        shares.groupby("technology").sum("asset").rename(technology="asset")
    )

    # Split demand over assets according to shares
    split_demand = demand * (grouped_shares / grouped_shares.sum("asset")).fillna(0)

    # Split unassigned demand equally over assets
    unassigned = (demand - split_demand.sum("asset")).clip(min=0)
    unassigned_per_asset = unassigned / grouped_shares.sizes["asset"]
    split_demand += unassigned_per_asset
    return split_demand


def unmet_demand(
    demand: xr.DataArray,
    capacity: xr.DataArray,
    technologies: xr.Dataset,
    timeslice_level: str | None = None,
):
    r"""Share of the demand that cannot be serviced by the existing assets.

    .. math::
        U[\mathcal{M}, \mathcal{A}] =
          \max(\mathcal{C} - P[\mathcal{M}, \mathcal{A}], 0)

    :math:`\max` operates element-wise, and indices have been dropped for simplicity.
    The resulting expression has the same indices as the consumption
    :math:`\mathcal{C}_{c, s}^r`.

    :math:`P` is the maximum production, given by
    :py:func:`muse.quantities.maximum_production`.
    """
    from muse.quantities import maximum_production

    # Check inputs
    assert "year" not in technologies.dims
    assert "year" not in capacity.dims
    assert "year" not in demand.dims

    # Calculate maximum production by existing assets
    produced = maximum_production(
        capacity=capacity,
        technologies=technologies,
        timeslice_level=timeslice_level,
    )

    # Total commodity production by summing over assets
    if "dst_region" in produced.dims:
        produced = produced.sum("asset").rename(dst_region="region")
    elif "region" in produced.coords and produced.region.dims:
        produced = produced.groupby("region").sum("asset")
    else:
        produced = produced.sum("asset")

    # Return unmet demand
    return (demand - produced).clip(min=0)


def new_consumption(
    capacity: xr.DataArray,
    demand: xr.DataArray,
    technologies: xr.Dataset,
    timeslice_level: str | None = None,
) -> xr.DataArray:
    r"""Computes share of the demand attributed to new agents.

    The new agents service the demand that can be attributed specifically to growth and
    that cannot be serviced by existing assets. In other words:

    .. math::
        N_{c, s}^r = \min\left(
            C_{c, s}^(y + \Delta y) - C_{c, s}^(y),
            C_{c, s}^(y + \Delta y)
                - P[\mathcal{M}(y + \Delta y), \mathcal{A}_{a, s}^r(y)]
        \right)

    Where :math:`P` the maximum production by existing assets, given by
    :py:func:`muse.quantities.maximum_production`.
    """
    # Validate inputs have matching years
    if not (
        len(demand.year) == len(capacity.year) == 2
        and (demand.year.values == capacity.year.values).all()
    ):
        raise ValueError("Capacity and demand must have matching years")

    current_year, investment_year = map(int, demand.year.values)

    # Calculate demand growth
    current_demand = demand.sel(year=current_year, drop=True)
    future_demand = demand.sel(year=investment_year, drop=True)
    new_demand = (future_demand - current_demand).clip(min=0)

    # If future capacity is higher than existing capacity, it's possible that
    # this might already be able to make up some of the increase in demand
    missing = unmet_demand(
        demand=future_demand,
        capacity=capacity.sel(year=investment_year),
        technologies=technologies,
        timeslice_level=timeslice_level,
    )

    # Return minimum of growth and unmet demand
    return np.minimum(new_demand, missing)


def new_and_retro_demands(
    capacity: xr.DataArray,
    demand: xr.DataArray,
    technologies: xr.Dataset,
    timeslice_level: str | None = None,
) -> xr.Dataset:
    """Splits demand into *new* and *retrofit* demand.

    The demand in the investment year is split three ways:

    #. the demand that can be serviced by the assets that will still be operational that
        year.
    #. the *new* demand is defined as the growth in consumption that cannot be serviced
        by existing assets in the current year, as computed in :py:func:`new_demand`.
    #. the retrofit demand is everything else.
    """
    # Validate inputs have matching years
    if not (
        len(demand.year) == len(capacity.year) == 2
        and (demand.year.values == capacity.year.values).all()
    ):
        raise ValueError("Capacity and demand must have matching years")

    investment_year = int(demand.year[1])

    if hasattr(capacity, "region") and capacity.region.dims == ():
        capacity["region"] = (
            "asset",
            [str(capacity.region.values)] * len(capacity.asset),
        )

    # Calculate new demand from growth
    new_demand = new_consumption(
        capacity=capacity,
        demand=demand,
        technologies=technologies,
        timeslice_level=timeslice_level,
    )

    # Calculate retrofit demand as remaining unmet demand
    retrofit_demand = (
        unmet_demand(
            demand=demand.sel(year=investment_year, drop=True),
            capacity=capacity.sel(year=investment_year),
            technologies=technologies,
            timeslice_level=timeslice_level,
        )
        - new_demand
    ).clip(0)

    return xr.Dataset({"new": new_demand, "retrofit": retrofit_demand})


def decommissioning_demand(
    technologies: xr.Dataset,
    capacity: xr.DataArray,
    timeslice_level: str | None = None,
) -> xr.DataArray:
    r"""Computes demand from process decommissioning.

    Let :math:`M_t^r(y)` be the retrofit demand, :math:`^{(s)}\mathcal{D}_t^r(y)` be the
    decommissioning demand at the level of the sector, and :math:`A^r_{t, \iota}(y)` be
    the assets owned by the agent. Then, the decommissioning demand for agent :math:`i`
    is :

    .. math::

        \mathcal{D}^{r, i}_{t, c}(y) =
            \sum_\iota \alpha_{t, \iota}^r \beta_{t, \iota, c}^r
                \left(A^{i, r}_{t, \iota}(y) - A^{i, r}_{t, \iota, c}(y + 1) \right)

    given the utilization factor :math:`\alpha_{t, \iota}` and the fixed output factor
    :math:`\beta_{t, \iota, c}`.

    Furthermore, decommissioning demand is non-zero only for end-use commodities.

    ncsearch-nohlsearch).. SeeAlso:
        :ref:`indices`, :ref:`quantities`,
        :py:func:`~muse.quantities.maximum_production`
        :py:func:`~muse.commodities.is_enduse`
    """
    from muse.quantities import maximum_production

    assert len(capacity.year) == 2
    assert "year" not in technologies.dims
    current_year, investment_year = map(int, capacity.year.values)

    # Calculate the decrease in capacity from the current year to future years
    capacity_decrease = capacity.sel(year=current_year) - capacity.sel(
        year=investment_year
    )

    # Calculate production associated with this capacity
    result = maximum_production(
        technologies,
        capacity_decrease,
        timeslice_level=timeslice_level,
    ).clip(min=0)
    assert "year" not in result.dims
    return result



"""MUSE model."""

from __future__ import annotations

import os
from contextlib import suppress
from importlib.metadata import PackageNotFoundError, version

with suppress(PackageNotFoundError):
    __version__ = version("MUSE_OS")


def _create_logger(color: bool = True):
    """Creates the main logger.

    Mainly a convenience function, so logging configuration can happen in one place.
    """
    import logging

    logger = logging.getLogger(name=__name__)
    formatter = "-- %(asctime)s - %(name)s - %(levelname)s\n%(message)s\n"

    if color:
        try:
            import coloredlogs

            coloredlogs.install(logger=logger, fmt=formatter)
        except ImportError:
            console = logging.StreamHandler()
            console.setFormatter(logging.Formatter(formatter, datefmt="%d-%m-%y %H:%M"))
            logger.addHandler(console)
    else:
        console = logging.StreamHandler()
        console.setFormatter(logging.Formatter(formatter, datefmt="%d-%m-%y %H:%M"))
        logger.addHandler(console)

    logger.setLevel(logging.DEBUG)

    return logger


def add_file_logger() -> None:
    """Adds a file logger to the main logger.

    The file logger is split into two files: one for INFO and DEBUG messages, and one
    for WARNING messages and above to avoid cluttering the main log file and highlight
    potential issues.
    """
    import logging
    from pathlib import Path

    from .defaults import DEFAULT_OUTPUT_DIRECTORY

    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )

    DEFAULT_OUTPUT_DIRECTORY.mkdir(parents=True, exist_ok=True)

    # Sets the warning log, for warnings and above
    warning_file = Path(DEFAULT_OUTPUT_DIRECTORY) / "muse_warning.log"
    if warning_file.exists():
        warning_file.unlink()

    warning_file_handler = logging.FileHandler(warning_file)
    warning_file_handler.setLevel(logging.WARNING)
    warning_file_handler.setFormatter(formatter)
    warning_file_handler.filters = [lambda record: record.levelno > logging.INFO]

    logging.getLogger("muse").addHandler(warning_file_handler)

    # Sets the info log, for debug and info only
    info_file = Path(DEFAULT_OUTPUT_DIRECTORY) / "muse_info.log"
    if info_file.exists():
        info_file.unlink()

    info_file_handler = logging.FileHandler(info_file)
    info_file_handler.setLevel(logging.DEBUG)
    info_file_handler.setFormatter(formatter)
    info_file_handler.filters = [lambda record: record.levelno <= logging.INFO]

    logging.getLogger("muse").addHandler(info_file_handler)


logger = _create_logger(os.environ.get("MUSE_COLOR_LOG") != "False")
""" Main logger """

__all__ = [
    "Agent",
    "create_agent",
    "decisions",
    "demand_share",
    "filters",
    "hooks",
    "interactions",
    "investments",
    "objectives",
    "outputs",
    "read_agent_parameters",
    "read_global_commodities",
    "read_initial_capacity",
    "read_io_technodata",
    "read_macro_drivers",
    "read_settings",
    "read_technodictionary",
    "read_technologies",
    "read_timeslice_shares",
    "sectors",
]



"""Various ways and means to compute production.

Production is the amount of commodities produced by an asset. However, depending on the
context, it could be computed several ways. For  instance, it can be obtained straight
from the capacity of the asset. Or it can be obtained by matching for the same
commodities with a set of assets.

Production methods can be registered via the :py:func:`@register_production
<register_production>` production decorator.  Registering a function makes the function
accessible from MUSE's input file. Production methods are not expected to modify their
arguments. Furthermore they should conform the
following signatures:

.. code-block:: python

    @register_production
    def production(
        market: xr.Dataset, capacity: xr.DataArray, technologies: xr.Dataset, **kwargs
    ) -> xr.DataArray:
        pass


Arguments:
    market: Market, including demand and prices.
    capacity: The capacity of each asset within a market.
    technologies: A dataset characterising the technologies of the same assets.
    **kwargs: Any number of keyword arguments

Returns:
    A `xr.DataArray` with the amount produced for each good from each asset.
"""

from __future__ import annotations

__all__ = [
    "PRODUCTION_SIGNATURE",
    "factory",
    "maximum_production",
    "register_production",
    "supply",
]

from collections.abc import MutableMapping
from typing import Callable

import xarray as xr

from muse.registration import registrator

"""Production signature."""
PRODUCTION_SIGNATURE = Callable[
    [xr.Dataset, xr.DataArray, xr.Dataset, str], xr.DataArray
]

"""Dictionary of production methods. """
PRODUCTION_METHODS: MutableMapping[str, PRODUCTION_SIGNATURE] = {}


@registrator(registry=PRODUCTION_METHODS, loglevel="info")
def register_production(function: PRODUCTION_SIGNATURE):
    """Decorator to register a function as a production method.

    .. seealso::

        :py:mod:`muse.production`
    """
    return function


def factory(name) -> PRODUCTION_SIGNATURE:
    from muse.production import PRODUCTION_METHODS

    return PRODUCTION_METHODS[name]


@register_production(name=("max", "maximum"))
def maximum_production(
    market: xr.Dataset,
    capacity: xr.DataArray,
    technologies: xr.Dataset,
    timeslice_level: str | None = None,
) -> xr.DataArray:
    """Production when running at full capacity.

    *Full capacity* is limited by the utilization factor. For more details, see
    :py:func:`muse.quantities.maximum_production`.
    """
    from muse.quantities import maximum_production

    return maximum_production(technologies, capacity, timeslice_level)


@register_production(name=("share", "shares"))
def supply(
    market: xr.Dataset,
    capacity: xr.DataArray,
    technologies: xr.Dataset,
    timeslice_level: str | None = None,
) -> xr.DataArray:
    """Service current demand equally from all assets.

    "Equally" means that equivalent technologies are used to the same percentage of
    their respective capacity.
    """
    from muse.quantities import supply

    return supply(
        capacity, market.consumption, technologies, timeslice_level=timeslice_level
    )



"""Sector with preset behaviours."""

from __future__ import annotations

from typing import Any

from xarray import DataArray, Dataset

from muse.sectors.register import AbstractSector, register_sector


@register_sector(name=("preset", "presets"))
class PresetSector(AbstractSector):  # type: ignore
    """Sector with outcomes fixed from the start."""

    @classmethod
    def factory(cls, name: str, settings: Any) -> PresetSector:
        """Constructs a PresetSectors from input data."""
        from muse.readers.toml import read_presets_sector

        presets = read_presets_sector(settings, name)
        sector_conf = getattr(settings.sectors, name)
        interpolation_mode = getattr(sector_conf, "interpolation_mode", "linear")
        return cls(presets, interpolation_mode=interpolation_mode, name=name)

    def __init__(
        self,
        presets: Dataset,
        interpolation_mode: str = "linear",
        name: str = "preset",
    ):
        super().__init__()

        self.presets: Dataset = presets
        """Market across time and space."""
        self.interpolation_mode: str = interpolation_mode
        """Interpolation method"""
        self.name = name
        """Name by which to identify a sector"""
        self.commodities: list[str] = []

    def next(self, mca_market: Dataset) -> Dataset:
        """Advance sector by one time period."""
        presets = self.presets.sel(region=mca_market.region)
        supply = self._interpolate(presets.supply, mca_market.year)
        consumption = self._interpolate(presets.consumption, mca_market.year)
        costs = self._interpolate(presets.costs, mca_market.year)

        result = Dataset({"supply": supply, "consumption": consumption, "costs": costs})
        assert isinstance(result, Dataset)
        return result

    def _interpolate(self, data: DataArray, years: DataArray) -> DataArray:
        return data.interp(year=years, method=self.interpolation_mode).ffill("year")
